#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import urllib.request
import string
import os
import datetime
import dbm.dumb

from ..mkExceptions import JournalDataBaseError

abrevDBname = "j9Abbreviations"

manualDBname = "manualj9Abbreviations"

def j9urlGenerator(nameDict = False):
    """How to get all the urls for the WOS Journal Title Abbreviations. Each is varies by only a few characters. These are the currently in use urls they may change.

    They are of the form:

    > "https://images.webofknowledge.com/images/help/WOS/{VAL}_abrvjt.html"
    > Where {VAL} is a capital letter or the string "0-9"

    # Returns

    `list[str]`

    > A list of all the url's strings
    """
    start = "https://images.webofknowledge.com/images/help/WOS/"
    end = "_abrvjt.html"
    if nameDict:
        urls = {"0-9" : start + "0-9" + end}
        for c in string.ascii_uppercase:
            urls[c] = start + c + end
    else:
        urls = [start + "0-9" + end]
        for c in string.ascii_uppercase:
            urls.append(start + c + end)
    return urls

def _j9SaveCurrent(sDir = '.'):
    """Downloads and saves all the webpages

    For Backend
    """
    dname = os.path.normpath(sDir + '/' +  datetime.datetime.now().strftime("%Y-%m-%d_J9_AbbreviationDocs"))
    if not os.path.isdir(dname):
        os.mkdir(dname)
        os.chdir(dname)
    else:
        os.chdir(dname)
    for urlID, urlString in j9urlGenerator(nameDict = True).items():
        fname = "{}_abrvjt.html".format(urlID)
        f = open(fname, 'wb')
        f.write(urllib.request.urlopen(urlString).read())

def _getDict(j9Page):
    """Parses a Journal Title Abbreviations page

    Note the pages are not well formatted html as the <DT> tags are not closes so html parses (Beautiful Soup) do not work. This is a simple parser that only works on the webpages and may fail if they are changed

    For Backend
    """
    slines = j9Page.read().decode('utf-8').split('\n')
    while slines.pop(0) != "<DL>":
        pass
    currentName = slines.pop(0).split('"></A><DT>')[1]
    currentTag = slines.pop(0).split("<B><DD>\t")[1]
    j9Dict = {}
    while True:
        try:
            j9Dict[currentTag].append(currentName)
        except KeyError:
            j9Dict[currentTag] = [currentName]
        try:
            currentName = slines.pop(0).split('</B><DT>')[1]
            currentTag = slines.pop(0).split("<B><DD>\t")[1]
        except IndexError:
            break
    return j9Dict

def _getCurrentj9Dict():
    """Downloads and parses all the webpages

    For Backend
    """
    urls = j9urlGenerator()
    j9Dict = {}
    for url in urls:
        d = _getDict(urllib.request.urlopen(url))
        if len(d) == 0:
            raise RuntimeError("Parsing failed, this is could require an update of the parser.")
        j9Dict.update(d)
    return j9Dict

def updatej9DB(dbname = abrevDBname, saveRawHTML = False):
    """Updates the database of Journal Title Abbreviations. Requires an internet connection. The data base is saved relative to the source file not the working directory.

    # Parameters

    _dbname_ : `optional [str]`

    > The name of the database file, default is "j9Abbreviations.db"

    _saveRawHTML_ : `optional [bool]`

    > Determines if the original HTML of the pages is stored, default `False`. If `True` they are saved in a directory inside j9Raws begining with todays date.
    """
    if saveRawHTML:
        rawDir = '{}/j9Raws'.format(os.path.dirname(__file__))
        if not os.path.isdir(rawDir):
            os.mkdir(rawDir)
        _j9SaveCurrent(sDir = rawDir)
    dbLoc = os.path.join(os.path.normpath(os.path.dirname(__file__)), dbname)
    try:
        with dbm.dumb.open(dbLoc, flag = 'c') as db:
            try:
                j9Dict = _getCurrentj9Dict()
            except urllib.error.URLError:
                raise urllib.error.URLError("Unable to access server, check your connection")
            for k, v in j9Dict.items():
                if k in db:
                    for jName in v:
                        if jName not in j9Dict[k]:
                            j9Dict[k] += '|' + jName
                else:
                    db[k] = '|'.join(v)
    except dbm.dumb.error as e:
        raise JournalDataBaseError("Something happened with the database of WOS journal names. To fix this you should delete the 1 to 3 files whose names start with {}. If this doesn't work (sorry), deleteing everything in '{}' and reinstalling metaknowledge should.\nThe error was '{}'".format(dbLoc, os.path.dirname(__file__), e))

def getj9dict(dbname = abrevDBname, manualDB = manualDBname, returnDict ='both'):
    """Returns the dictionary of journal abbreviations mapping to a list of the associated journal names. By default the local database is used. The database is in the file _dbname_ in the same directory as this source file

    # Parameters

    _dbname_ : `optional [str]`

    > The name of the downloaded database file, the default is determined at run time. It is recommended that this remain untouched.

    _manualDB_ : `optional [str]`

    > The name of the manually created database file, the default is determined at run time. It is recommended that this remain untouched.

    _returnDict_ : `optional [str]`

    > default `'both'`, can be used to get both databases or only one  with `'WOS'` or `'manual'`.
    """
    dbLoc = os.path.normpath(os.path.dirname(__file__))

    retDict = {}
    try:
        if returnDict == 'both' or returnDict == 'WOS':
            with dbm.dumb.open(dbLoc + '/{}'.format(dbname)) as db:
                if len(db) == 0:
                    raise JournalDataBaseError("J9 Database empty or missing, to regenerate it import and run metaknowledge.WOS.journalAbbreviations.updatej9DB().")
                for k, v in db.items():
                    retDict[k.decode('utf-8')] = v.decode('utf-8').split('|')
    except JournalDataBaseError:
        updatej9DB()
        return getj9dict(dbname = dbname, manualDB = manualDB, returnDict = returnDict)
    try:
        if returnDict == 'both' or returnDict == 'manual':
            if os.path.isfile(dbLoc + '/{}.dat'.format(manualDB)):
                with dbm.dumb.open(dbLoc + '/{}'.format(manualDB)) as db:
                    for k, v in db.items():
                        retDict[k.decode('utf-8')] = v.decode('utf-8').split('|')
            else:
                if returnDict == 'manual':
                    raise JournalDataBaseError("Manual J9 Database ({0}) missing, to create it run addToDB(dbname = {0})".format(manualDB))
    except JournalDataBaseError:
        updatej9DB(dbname = manualDB)
        return getj9dict(dbname = dbname, manualDB = manualDB, returnDict = returnDict)
    return retDict

def addToDB(abbr = None, dbname = manualDBname):
    """Adds _abbr_ to the database of journals. The database is kept separate from the one scraped from WOS, this supersedes it. The database by default is stored with the WOS one and the name is given by `metaknowledge.journalAbbreviations.manualDBname`. To create an empty database run **addToDB** without an _abbr_ argument.

    # Parameters

    _abbr_ : `optional [str or dict[str : str]]`

    > The journal abbreviation to be added to the database, it can either be a single string in which case that string will be added with its self as the full name, or a dict can be given with the abbreviations as keys and their names as strings, use pipes (`'|'`) to separate multiple names. Note, if the empty string is given as a name the abbreviation will be considered manually __excluded__, i.e. having excludeFromDB() run on it.

    _dbname_ : `optional [str]`

    > The name of the database file, default is `metaknowledge.journalAbbreviations.manualDBname`.
    """
    dbLoc = os.path.normpath(os.path.dirname(__file__))
    with dbm.dumb.open(dbLoc + '/' + dbname) as db:
        if isinstance(abbr, str):
            db[abbr] = abbr
        elif isinstance(abbr, dict):
            try:
                db.update(abbr)
            except TypeError:
                raise TypeError("The keys and values of abbr must be strings.")
        elif abbr is None:
            pass
        else:
            raise TypeError("abbr must be a str or dict.")

def excludeFromDB(abbr = None, dbname = manualDBname):
    """Marks _abbr_ to be excluded the database of journals. The database is kept separate from the one scraped from WOS, this supersedes it. The database by default is stored with the WOS one and the name is given by `metaknowledge.journalAbbreviations.manualDBname`. To create an empty database run [addToDB()](#metaknowledge.journalAbbreviations.backend.addToDB) without an _abbr_ argument.

    # Parameters

    _abbr_ : `optional [str or tuple[str] or list[str]`

    > The journal abbreviation to be excluded from the database, it can either be a single string in which case that string will be exclude or a list/tuple of strings can be given with the abbreviations.

    _dbname_ : `optional [str]`

    > The name of the database file, default is `metaknowledge.journalAbbreviations.manualDBname`.
    """
    dbLoc = os.path.normpath(os.path.dirname(__file__))
    with dbm.dumb.open(dbLoc + '/' + dbname) as db:
        if isinstance(abbr, str):
            db[abbr] = ''
        elif isinstance(abbr, list) or isinstance(abbr, tuple):
            try:
                db.update({k : '' for k in abbr})
            except TypeError:
                raise TypeError("The keys and values of abbr must be strings.")
        elif abbr is None:
            pass
        else:
            raise TypeError("abbr must be a str, list or tuple.")



singleLineEntries = {
    #The inverse cannot be done as new tags may be added that occupy mutiple line
    'Author',
    'Title',
    'Advisor',
    'Name',
    'Committee member',
    'Copyright',
    'Country of publication',
    'Database',
    'Degree',
    'Degree date',
    'Department',
    'Dissertation/thesis number',
    'Document type',
    'Language',
    'Number of pages',
    'Place of publication',
    'ProQuest document ID',
    'Publication year',
    'School code',
    'Source',
    'Source type',
    'University location',
    'University/institution',
    'ISBN',
    'Publication subject',
}

def proQuestSubject(value):
    return value[0].split('; ')

def proQuestIdentifier_Keyword(value):
    return value[0].split(', ')

def proQuestClassification(value):
    return [tuple(s.split(': ')) for s in value[0].split('; ')]

customTags = {
    'Classification' : proQuestClassification,
    'Identifier / keyword' : proQuestIdentifier_Keyword,
    'Subject' : proQuestSubject
}

def proQuestTagToFunc(tag):
    """Takes a tag string, _tag_, and returns the processing function for its data. If their is not a predefined function returns the identity function (`lambda x : x`).

    # Parameters

    _tag_ : `str`

    > The requested tag

    # Returns

    `function`

    > A function to process the tag's data
    """
    if tag in singleLineEntries:
        return lambda x : x[0]
    elif tag in customTags:
        return customTags[tag]
    else:
        return lambda x : x



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
"""_metaknowledge_ is a Python3 package that simplifies bibliometric and computational analysis of Web of Science data.

# Example

To load the data from files and make a network:

    >>> import metaknowledge as mk
    >>> RC = mk.RecordCollection("records/")
    >>> print(RC)
    Collection of 33 records
    >>> G = RC.coCiteNetwork(nodeType = 'journal')
    Done making a co-citation network of files-from-records                 1.1s
    >>> print(len(G.nodes()))
    223
    >>> mk.writeGraph(G, "Cocitation-Network-of-Journals")

There is also a simple command line program called `metaknowledge` that comes with the package. It allows for creating networks without any need to know Python. More information about it can be found at [networkslab.org/metaknowledge/cli]({{ site.baseurl }}/cli)

# Overview

This package can read the files downloaded from the Thomson Reuters' [Web of Science](https://webofknowledge.com) (_WOS_), Elsevier's [Scopus](https://www.scopus.com/), [ProQuest](www.proquest.com/) and Medline files from [PubMed](www.ncbi.nlm.nih.gov/pubmed). These files contain entries on the metadata of scientific records, such as authors, title, and citations. _metaknowledge_ can also read grants from various organizations including _NSF_ and _NSERC_ which are handled similarly to records.

The [metaknowledge.RecordCollection](./documentation/classes/RecordCollection.html#recordcollection-collectionwithids) class can take a path to one or more of these files load and parse them. The object is the main way for work to be done on multiple records. For each individual record it creates an instance of the [metaknowledge.Record](./documentation/classes/Record.html#record-mapping-hashable) class that contains the results of the parsing of the record.

The files read by _metaknowledge_ are a databases containing a series of tags (implicitly or explicitly), e.g. `'TI'` is the title for WOS. Each tag has one or more values and metaknowledge can read them and extract useful information. As the tags differ between providers a small set of values can be accessed by special tags, the tags are listed in `commonRecordFields`. These special tags can act on the whole `Record` and as such may contain information provided by any number of other tags.

Citations are handled by a special [Citation](./documentation/classes/Citation.html#module-metaknowledge.citation) class. This class can parse the citations given by _WOS_ and journals cited by _Scopus_ and allows for better comparisons when they are used in graphs.

Note for those reading the docstrings metaknowledge's docs are written in markdown and are processed to produce the documentation found at [metaknowledge.readthedocs.io](https://metaknowledge.readthedocs.io/en/latest/), but you should have no problem reading them from the help function.
"""

from .constants import VERBOSE_MODE, __version__, commonRecordFields, FAST_CITES
from .mkExceptions import BadCitation, BadGrant, BadInputFile, BadProQuestFile, BadProQuestRecord, BadPubmedFile, BadPubmedRecord, BadRecord, BadWOSFile, BadWOSRecord, CollectionTypeError, GrantCollectionException, RCTypeError, RCValueError, RecordsNotCompatible, UnknownFile, cacheError, mkException, TagError, BadScopusRecord

from .graphHelpers import writeEdgeList, writeNodeAttributeFile, writeGraph, readGraph, dropEdges, dropNodesByDegree, dropNodesByCount, mergeGraphs, graphStats, writeTnetFile
from .diffusion import diffusionGraph, diffusionCount, diffusionAddCountsFromSource

from .citation import Citation, filterNonJournals
from .mkCollection import Collection, CollectionWithIDs
from .mkRecord import Record, ExtendedRecord

from .grantCollection import GrantCollection
from .grants import NSERCGrant, CIHRGrant, MedlineGrant, NSFGrant, Grant, FallbackGrant

from .recordCollection import RecordCollection
from .WOS import WOSRecord
from .medline import MedlineRecord
from .proquest import ProQuestRecord
from .scopus import ScopusRecord

from .journalAbbreviations.backend import updatej9DB
from .genders.nameGender import downloadData

def downloadExtras():
    """Downloads all the external files used by metaknowledge. This will overwrite exiting files
    """
    print("Downloading journal abbreviations data")
    updatej9DB()
    print("Downloading gender name data")
    downloadData()



scopusSpecialTagToFunc = {
    'year' : lambda R : R['Year'],
    'volume' : lambda R : R['Volume'],
    'beginningPage' : lambda R : R['Page start'],
    #'DOI' : lambda R : R['DOI'], Causese recursion errors if not commented out
    #'address' : lambda R : R[''],
    'j9' : lambda R : R['Abbreviated Source Title'],
    'citations' : lambda R : R['References'],
    #'grants' : lambda R : R['References'],
    'selfCitation' : lambda R : R.createCitation(),
    'authorsShort' : lambda R : R['Authors'],
    'authorsFull' : lambda R : R['Authors'],
    'title' : lambda R : R['Title'],
    'journal' : lambda R : R['Source title'],
    'keywords' : lambda R : R['Index Keywords'],
    'abstract' : lambda R : R['Abstract'],
    'id' : lambda R : R['EID'],
}



import copy
import pickle
import os
import os.path
import csv
try:
    import collections.abc
except ImportError:
    import collections
    collections.abc = collections

import networkx as nx

from .progressBar import _ProgressBar

from .RCglimpse import _glimpse

from .constants import __version__

from .mkExceptions import CollectionTypeError, cacheError, TagError, mkException

import metaknowledge

class Collection(collections.abc.MutableSet, collections.abc.Hashable):
    """A named hashable set with some error reporting.

    `Collections` have all the methods of builtin `sets` as well as error reporting with _bad_ and _error_, and control over the contained items with _allowedTypes_ and _collectedTypes_.

    # Customizations

    When created _name_ should be a string that allows users to easily determine the source of the `Collection`

    When created the you must provided a set of types, _allowedTypes_, when new items are added they will be checked and if they are not instances of any of the types an `CollectionTypeError` exception will be raised. The _collectedTypes_ set that is provided should be a set of only the types in the `Collection`.

    If any of the elements in the `Collection` are bad then _bad_ should be set to `True` and the `dict` _errors_ should map the item to it's exception.

    All of these customizations are managed when operations occur on the `Collection` and if 2 `Collections` are modified with one of the binary operators (`|`, `-`, etc) the `_collectedTypes` and `errors` attributes will be modified the same way. `name` will be updated to explain the operation(s) that occurred.

    \_\_Init\_\_

    As `Collection` is mostly meant to be base for other classes all but one of the arguments in the \_\_Init\_\_ are not optional and the optional one is not used.

    # Parameters

    _inSet_ : `set`

    > The objects to be contained

    _allowedTypes_ : `set[type]`

    > A set of types, `{object}` will allow virtually everything

    _collectedTypes_ : `set[type]`

    > The types (or supertypes) of the objects in _inSet_

    _name_ : `str`

    > The name of the `Collection`

    _bad_ : `bool`

    > If any of the elements are bad

    _errors_ : `dict[:Exception]`

    > A mapping from items to their errors

    _quietStart_ : `optional [bool]`

    > Default `False`, does nothing. This is here for use as a interface by subclasses
    """
    def __init__(self, inSet, allowedTypes, collectedTypes, name, bad, errors, quietStart = False):
        """Basically a collections.abc.MutableSet wrapper for a set with a bunch of extra record keeping attached."""
        self._collection = inSet
        self._allowedTypes = allowedTypes
        self._collectedTypes = collectedTypes

        self.name = name
        self.bad = bad
        self.errors = errors

    #Hashable method

    def __hash__(self):
        return hash(sum((hash(i) for i in self)))

    #Set methods

    def __le__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            return len(self) <= len(other)

    def __ge__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            return len(self) >= len(other)

    def __eq__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            return self._collection == other._collection

    def __len__(self):
        return len(self._collection)

    def __iter__(self):
        for i in self._collection:
            yield i

    def __contains__(self, item):
        return item in self._collection

    #Mutable Set methods

    def add(self, elem):
        """ Adds _elem_ to the collection.

        # Parameters

        _elem_ : `object`

        > The object to be added
        """
        if isinstance(elem, self._allowedTypes):
            self._collection.add(elem)
            self._collectedTypes.add(type(elem).__name__)
        else:
            raise CollectionTypeError("{} can only contain '{}', '{}' is not allowed.".format(type(self).__name__, self._allowedTypes, elem))

    def discard(self, elem):
        """Removes _elem_ from the collection, will not raise an Exception if _elem_ is missing

        # Parameters

        _elem_ : `object`

        > The object to be removed

        """
        return self._collection.discard(elem)

    def remove(self, elem):
        """Removes _elem_ from the collection, will raise a KeyError is _elem_ is missing

        # Parameters

        _elem_ : `object`

        > The object to be removed
        """
        try:
            return self._collection.remove(elem)
        except KeyError:
            raise KeyError("'{}' was not found in the {}: '{}'.".format(elem, type(self).__name__, self)) from None

    def clear(self):
        """"Removes all elements from the collection and resets the error handling
        """
        self.bad = False
        self.errors = {}
        self._collection.clear()

    def pop(self):
        """Removes a random element from the collection and returns it

        # Returns

        `object`

        > A random object from the collection
        """
        try:
            return self._collection.pop()
        except KeyError:
            raise KeyError("Nothing left in the {}: '{}'.".format(type(self).__name__, self)) from None

    def __ior__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            self._collection |= other._collection
            self._collectedTypes |= other._collectedTypes
            self.name = '{} |= {}'.format(self.name, other.name)
            if other.bad or self.bad:
                self.bad = True
                self.errors.update(other.errors)
            return self

    def __iand__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            self._collection &= other._collection
            self._collectedTypes |= other._collectedTypes
            self.name = '{} &= {}'.format(self.name, other.name)
            if other.bad or self.bad:
                self.bad = True
                self.errors.update(other.errors)
            return self

    def __ixor__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            self._collection ^= other._collection
            self._collectedTypes |= other._collectedTypes
            self.name = '{} ^= {}'.format(self.name, other.name)
            if other.bad or self.bad:
                self.bad = True
                self.errors.update(other.errors)
            return self

    def __isub__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            self._collection -= other._collection
            self._collectedTypes |= other._collectedTypes
            self.name = '{} -= {}'.format(self.name, other.name)
            if other.bad or self.bad:
                self.bad = True
                self.errors.update(other.errors)
            return self

    #These are provided by the above
    #but don't work right unless they are custom written

    def __or__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            retCollection = type(self)(self._collection | other._collection, name = '{} | {}'.format(self.name, other.name), quietStart = True)
            if other.bad or self.bad:
                retCollection.bad = True
                retCollection.errors.update(other.errors)
            return retCollection

    def __and__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            retCollection = type(self)(self._collection & other._collection, name = '{} & {}'.format(self.name, other.name), quietStart = True)
            if other.bad or self.bad:
                retCollection.bad = True
                retCollection.errors.update(other.errors)
            return retCollection

    def __sub__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            retCollection = type(self)(self._collection - other._collection, name = '{} - {}'.format(self.name, other.name), quietStart = True)
            if other.bad or self.bad:
                retCollection.bad = True
                retCollection.errors.update(other.errors)
            return retCollection

    def __xor__(self, other):
        if type(self) != type(other):
            return NotImplemented
        else:
            retCollection = type(self)(self._collection ^ other._collection, name = '{} ^ {}'.format(self.name, other.name), quietStart = True)
            if other.bad or self.bad:
                retCollection.bad = True
                retCollection.errors.update(other.errors)
            return retCollection

    def __repr__(self):
        return "<metaknowledge.{} object {}>".format(type(self).__name__, self.name)

    def __str__(self):
        return "{}({})".format(type(self).__name__, self.name)

    def copy(self):
        """Creates a shallow copy of the collection

        # Returns

        `Collection`

        > A copy of the `Collection`
        """
        collectedCopy = copy.copy(self)
        collectedCopy._collection = copy.copy(collectedCopy._collection)
        self._collectedTypes = copy.copy(self._collectedTypes)
        self._allowedTypes = copy.copy(self._allowedTypes)
        collectedCopy.errors = copy.copy(collectedCopy.errors)
        return collectedCopy

    def peek(self):
        """returns a random element from the collection. If ran twice the same element will usually be returned

        # Returns

        `object`

        > A random object from the collection
        """
        if len(self._collection) > 0:
            return self._collection.__iter__().__next__()
        else:
            return None

    def chunk(self, maxSize):
        """Splits the `Collection` into _maxSize_ size or smaller `Collections`

        # Parameters

        _maxSize_ : `int`

        > The maximum number of elements in a retuned `Collection`


        # Returns

        `list [Collection]`

        > A list of `Collections` that if all merged (`|` operator) would create the original
        """
        chunks = []
        currentSize = maxSize + 1
        for i in self:
            if currentSize >= maxSize:
                currentSize = 0
                chunks.append(type(self)({i}, name = 'Chunk-{}-of-{}'.format(len(chunks), self.name), quietStart = True))
            else:
                chunks[-1].add(i)
            currentSize += 1
        return chunks

    def split(self, maxSize):
        """Destructively, splits the `Collection` into _maxSize_ size or smaller `Collections`. The source `Collection` will be empty after this operation

        # Parameters

        _maxSize_ : `int`

        > The maximum number of elements in a retuned `Collection`

        # Returns

        `list [Collection]`

        > A list of `Collections` that if all merged (`|` operator) would create the original
        """
        chunks = []
        currentSize = maxSize + 1
        try:
            while True:
                if currentSize >= maxSize:
                    currentSize = 0
                    chunks.append(type(self)({self.pop()}, name = 'Chunk-{}-of-{}'.format(len(chunks), self.name), quietStart = True))
                else:
                    chunks[-1].add(self.pop())
                currentSize += 1
        except KeyError:
            self.clear()
            self.name = 'Emptied-{}'.format(self.name)
        return chunks

    def _loadFromCache(self, cacheName, flist, name, extension):
        def loadCache(cacheFile, flist, rcName, fileExtensions):
            with open(cacheFile, 'rb') as f:
                try:
                    dat, RC = pickle.load(f)
                except pickle.PickleError as e:
                    raise cacheError("pickle Error: {}".format(e))
            if dat["metaknowledge Version"] != __version__:
                raise cacheError("mk version mismatch")
            if dat["Collection Name"] != rcName:
                raise cacheError("Name mismatch")
            if dat["File Extension"] != fileExtensions:
                raise cacheError("Extension mismatch")
            if len(flist) != len(dat["File dict"]):
                raise cacheError("File number mismatch")
            flist = flist.copy()
            while len(flist) > 0:
                workingFile = flist.pop()
                if os.stat(workingFile).st_mtime != dat["File dict"][workingFile]:
                    raise cacheError("File modification mismatch")
            return RC.__dict__

        if os.path.isfile(cacheName):
            try:
                self.__dict__ = loadCache(cacheName, flist, name, extension)
            except (cacheError, KeyError, FileNotFoundError):
                os.remove(cacheName)
                return False
            else:
                return True

    def _createCache(self, cacheFile, flist, name, extension):
        dat = {
            "metaknowledge Version" : __version__,
            "File dict" : {},
            "Collection Name" : name,
            "File Extension" : extension,
        }
        for fileName in flist:
            dat["File dict"][fileName] =  os.stat(fileName).st_mtime
        with open(cacheFile, 'wb') as f:
            pickle.dump((dat, self), f)

class CollectionWithIDs(Collection):
    """A [Collection](./Collection.html#metaknowledge.Collection) with a few extra methods that assume all the contained items have an id attribute and a bad attribute, e.g. [Records](./Record.html#metaknowledge.Record) or [Grants](./Grant.html#metaknowledge.grants.Grant).

    \_\_Init\_\_

    As `CollectionWithIDs` is mostly meant to be base for other classes all but one of the arguments in the `__init__` are not optional and the optional one is not used. The `__init__()` function is the same as a [Collection](./Collection.html#metaknowledge.Collection).
    """
    def __init__(self, inSet, allowedTypes, collectedTypes, name, bad, errors, quietStart = False):

        Collection.__init__(self, inSet, allowedTypes, collectedTypes, name, bad, errors, quietStart = quietStart)

    def containsID(self, idVal):
        """Checks if the collected items contains the give _idVal_

        # Parameters

        _idVal_ : `str`

        > The queried id string

        # Returns

        `bool`

        > `True` if the item is in the collection
        """
        for i in self:
            if i.id == idVal:
                return True
        return False

    def discardID(self, idVal):
        """Checks if the collected items contains the give _idVal_ and discards it if it is found, will not raise an exception if item is not found

        # Parameters

        _idVal_ : `str`

        > The discarded id string
        """
        for i in self:
            if i.id == idVal:
                self._collection.discard(i)
                return

    def removeID(self, idVal):
        """Checks if the collected items contains the give _idVal_ and removes it if it is found, will raise a `KeyError` if item is not found

        # Parameters

        _idVal_ : `str`

        > The removed id string
        """
        for i in self:
            if i.id == idVal:
                self._collection.remove(i)
                return
        raise KeyError("A Record with the ID '{}' was not found in the RecordCollection: '{}'.".format(idVal, self))

    def getID(self, idVal):
        """Looks up an item with _idVal_ and returns it if it is found, returns `None` if it does not find the item

        # Parameters

        _idVal_ : `str`

        > The requested item's id string

        # Returns

        `object`

        > The requested object or `None`
        """
        for i in self:
            if i.id == idVal:
                return i
        return None

    def badEntries(self):
        """Creates a new collection of the same type with only the bad entries

        # Returns

        `CollectionWithIDs`

        > A collection of only the bad entries
        """
        badEntries = set()
        for i in self:
            if i.bad:
                badEntries.add(i)
        return type(self)(badEntries, quietStart = True)

    def dropBadEntries(self):
        """Removes all the bad entries from the collection
        """
        self._collection = set((i for i in self if not i.bad))
        self.bad = False
        self.errors = {}

    def tags(self):
        """Creates a list of all the tags of the contained items

        # Returns

        `list [str]`

        > A list of all the tags
        """
        tags = set()
        for i in self:
            tags |= set(i.keys())
        return tags

    def glimpse(self, *tags, compact = False):
        """Creates a printable table with the most frequently occurring values of each of the requested _tags_, or if none are provided the top authors, journals and citations. The table will be as wide and as tall as the terminal (or 80x24 if there is no terminal) so `print(RC.glimpse())`should always create a nice looking table. Below is a table created from some of the testing files:

        ```
        >>> print(RC.glimpse())
        +RecordCollection glimpse made at: 2016-01-01 12:00:00++++++++++++++++++++++++++
        |33 Records from testFile++++++++++++++++++++++++++++++++++++++++++++++++++++++|
        |Columns are ranked by num. of occurrences and are independent of one another++|
        |-------Top Authors--------+------Top Journals-------+--------Top Cited--------|
        |1                Girard, S|1 CANADIAN JOURNAL OF PH.|1 LEVY Y, 1975, OPT COMM.|
        |1                Gilles, H|1 JOURNAL OF THE OPTICAL.|2 GOOS F, 1947, ANN PHYS.|
        |2                IMBERT, C|2          APPLIED OPTICS|3 LOTSCH HKV, 1970, OPTI.|
        |2                Pillon, F|2   OPTICS COMMUNICATIONS|4 RENARD RH, 1964, J OPT.|
        |3          BEAUREGARD, OCD|2 NUOVO CIMENTO DELLA SO.|5 IMBERT C, 1972, PHYS R.|
        |3               Laroche, M|2 JOURNAL OF THE OPTICAL.|6 ARTMANN K, 1948, ANN P.|
        |3                 HUARD, S|2 JOURNAL OF THE OPTICAL.|6 COSTADEB.O, 1973, PHYS.|
        |4                  PURI, A|2 NOUVELLE REVUE D OPTIQ.|6 ROOSEN G, 1973, CR ACA.|
        |4               COSTADEB.O|3 PHYSICS REPORTS-REVIEW.|7 Imbert C., 1972, Nouve.|
        |4           PATTANAYAK, DN|3 PHYSICAL REVIEW LETTERS|8 HOROWITZ BR, 1971, J O.|
        |4           Gazibegovic, A|3 USPEKHI FIZICHESKIKH N.|8 BRETENAKER F, 1992, PH.|
        |4                ROOSEN, G|3 APPLIED PHYSICS B-LASE.|8 SCHILLIN.H, 1965, ANN .|
        |4               BIRMAN, JL|3 AEU-INTERNATIONAL JOUR.|8 FEDOROV FI, 1955, DOKL.|
        |4                Kaiser, R|3 COMPTES RENDUS HEBDOMA.|8 MAZET A, 1971, CR ACAD.|
        |5                  LEVY, Y|3 CHINESE PHYSICS LETTERS|9 IMBERT C, 1972, CR ACA.|
        |5              BEAUREGA.OC|3       PHYSICAL REVIEW B|9 LOTSCH HKV, 1971, OPTI.|
        |5               PAVLOV, VI|3 LETTERE AL NUOVO CIMEN.|9 ASHBY N, 1973, PHYS RE.|
        |5                BREVIK, I|3 PROGRESS IN QUANTUM EL.|9 BOULWARE DG, 1973, PHY.|
        >>>
        ```

        # Parameters

        _tags_ : `str, str, ...`

        > Any number of tag strings to be made into columns in the output table

        # Returns

        `str`

        > A string containing the table
        """
        return _glimpse(self, *tags, compact = compact)

    def rankedSeries(self, tag, outputFile = None, giveCounts = True, giveRanks = False, greatestFirst = True, pandasMode = True, limitTo = None):
        """Creates an pandas dict of the ordered list of all the values of _tag_, with and ranked by their number of occurrences. A list can also be returned with the the counts or ranks added or it can be written to a file.

        # Parameters

        _tag_ : `str`

        > The tag to be ranked

        _outputFile_ : `optional str`

        > A file path to write a csv with 2 columns, one the tag values the other their counts

        _giveCounts_ : `optional bool`

        > Default `True`, if `True` the retuned list will be composed of tuples the first values being the tag value and the second their counts. This supersedes _giveRanks_.

        _giveRanks_ : `optional bool`

        > Default `False`, if `True` and _giveCounts_ is `False`, the retuned list will be composed of tuples the first values being the tag value and the second their ranks. This is superseded by _giveCounts_.

        _greatestFirst_ : `optional bool`

        > Default `True`, if `True` the returned list will be ordered with the highest ranked value first, otherwise the lowest ranked will be first.

        _pandasMode_ : `optional bool`

        > Default `True`, if `True` a `dict` ready for pandas will be returned, otherwise a list

        _limitTo_ : `optional list[values]`

        > Default `None`, if a list is provided only those values in the list will be counted or returned

        # Returns

        `dict[str:list[value]] or list[str]`

        > A `dict` or `list` will be returned depending on if _pandasMode_ is `True`
        """
        if giveRanks and giveCounts:
            raise mkException("rankedSeries cannot return counts and ranks only one of giveRanks or giveCounts can be True.")
        seriesDict = {}
        for R in self:
            #This should be faster than using get, since get is a wrapper for __getitem__
            try:
                val = R[tag]
            except KeyError:
                continue
            if not isinstance(val, list):
                val = [val]
            for entry in val:
                if limitTo and entry not in limitTo:
                    continue
                if entry in seriesDict:
                    seriesDict[entry] += 1
                else:
                    seriesDict[entry] = 1
        seriesList = sorted(seriesDict.items(), key = lambda x: x[1], reverse = greatestFirst)
        if outputFile is not None:
            with open(outputFile, 'w') as f:
                writer = csv.writer(f, dialect = 'excel')
                writer.writerow((str(tag), 'count'))
                writer.writerows(seriesList)
        if giveCounts and not pandasMode:
            return seriesList
        elif giveRanks or pandasMode:
            if not greatestFirst:
                seriesList.reverse()
            currentRank = 1
            retList = []
            panDict = {'entry' : [], 'count' : [], 'rank' : []}
            try:
                currentCount = seriesList[0][1]
            except IndexError:
                #Empty series so no need to loop
                pass
            else:
                for valString, count in seriesList:
                    if currentCount > count:
                        currentRank += 1
                        currentCount = count
                    if pandasMode:
                        panDict['entry'].append(valString)
                        panDict['count'].append(count)
                        panDict['rank'].append(currentRank)
                    else:
                        retList.append((valString, currentRank))
            if not greatestFirst:
                retList.reverse()
            if pandasMode:
                return panDict
            else:
                return retList
        else:
            return [e for e,c in seriesList]

    def timeSeries(self, tag = None, outputFile = None, giveYears = True, greatestFirst = True, limitTo = False, pandasMode = True):
        """Creates an pandas dict of the ordered list of all the values of _tag_, with and ranked by the year the occurred in, multiple year occurrences will create multiple entries. A list can also be returned with the the counts or years added or it can be written to a file.

        If no _tag_ is given the `Records` in the collection will be used

        # Parameters

        _tag_ : `optional str`

        > Default `None`, if provided the tag will be ordered

        _outputFile_ : `optional str`

        > A file path to write a csv with 2 columns, one the tag values the other their years

        _giveYears_ : `optional bool`

        > Default `True`, if `True` the retuned list will be composed of tuples the first values being the tag value and the second their years.

        _greatestFirst_ : `optional bool`

        > Default `True`, if `True` the returned list will be ordered with the highest years first, otherwise the lowest years will be first.

        _pandasMode_ : `optional bool`

        > Default `True`, if `True` a `dict` ready for pandas will be returned, otherwise a list

        _limitTo_ : `optional list[values]`

        > Default `None`, if a list is provided only those values in the list will be counted or returned

        # Returns

        `dict[str:list[value]] or list[str]`

        > A `dict` or `list` will be returned depending on if _pandasMode_ is `True`
        """
        seriesDict = {}
        for R in self:
            #This should be faster than using get, since get is a wrapper for __getitem__
            try:
                year = R['year']
            except KeyError:
                continue
            if tag is None:
                seriesDict[R] = {year : 1}
            else:
                try:
                    val = R[tag]
                except KeyError:
                    continue
                if not isinstance(val, list):
                    val = [val]
                for entry in val:
                    if limitTo and entry not in limitTo:
                        continue
                    if entry in seriesDict:
                        try:
                            seriesDict[entry][year] += 1
                        except KeyError:
                            seriesDict[entry][year] = 1
                    else:
                        seriesDict[entry] = {year : 1}
        seriesList = []
        for e, yd in seriesDict.items():
            seriesList += [(e, y) for y in yd.keys()]
        seriesList = sorted(seriesList, key = lambda x: x[1], reverse = greatestFirst)
        if outputFile is not None:
            with open(outputFile, 'w') as f:
                writer = csv.writer(f, dialect = 'excel')
                writer.writerow((str(tag), 'years'))
                writer.writerows(((k,'|'.join((str(y) for y in v))) for k,v in seriesDict.items()))
        if pandasMode:
            panDict = {'entry' : [], 'count' : [], 'year' : []}
            for entry, year in seriesList:
                panDict['entry'].append(entry)
                panDict['year'].append(year)
                panDict['count'].append(seriesDict[entry][year])
            return panDict
        elif giveYears:
            return seriesList
        else:
            return [e for e,c in seriesList]

    def cooccurrenceCounts(self, keyTag, *countedTags):
        """Counts the number of times values from any of the _countedTags_ occurs with _keyTag_. The counts are retuned as a dictionary with the values of _keyTag_ mapping to dictionaries with each of the _countedTags_ values mapping to thier counts.

        # Parameters

        _keyTag_ : `str`

        > The tag used as the key for the returned dictionary

        _*countedTags_ : `str, str, str, ...`

        > The tags used as the key for the returned dictionary's values

        # Returns

        `dict[str:dict[str:int]]`

        > The dictionary of counts
        """
        if not isinstance(keyTag, str):
            raise TagError("'{}' is not a string it cannot be used as a tag.".format(keyTag))
        if len(countedTags) < 1:
            TagError("You need to provide atleast one tag")
        for tag in countedTags:
            if not isinstance(tag, str):
                raise TagError("'{}' is not a string it cannot be used as a tag.".format(tag))
        occurenceDict = {}
        progArgs = (0, "Starting to count the co-occurrences of '{}' and' {}'".format(keyTag, "','".join(countedTags)))
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:

            for i, R in enumerate(self):
                PBar.updateVal(i / len(self), "Analyzing {}".format(R))
                keyVal = R.get(keyTag)
                if keyVal is None:
                    continue
                if not isinstance(keyVal, list):
                    keyVal = [keyVal]
                for key in keyVal:
                    if key not in occurenceDict:
                        occurenceDict[key] = {}
                for tag in countedTags:
                    tagval = R.get(tag)
                    if tagval is None:
                        continue
                    if not isinstance(tagval, list):
                        tagval = [tagval]
                    for val in tagval:
                        for key in keyVal:
                            try:
                                occurenceDict[key][val] += 1
                            except KeyError:
                                occurenceDict[key][val] = 1
            PBar.finish("Done extracting the co-occurrences of '{}' and '{}'".format(keyTag, "','".join(countedTags)))
        return occurenceDict

    def networkMultiLevel(self, *modes, nodeCount = True, edgeWeight = True, stemmer = None, edgeAttribute = None, nodeAttribute = None, _networkTypeString = 'n-level network'):
        """Creates a network of the objects found by any number of tags _modes_, with edges between all co-occurring values. IF you only want edges between co-occurring values from different tags use [networkMultiMode()](#metaknowledge.CollectionWithIDs.networkMultiMode).

        A **networkMultiLevel**() looks are each entry in the collection and extracts its values for the tag given by each of the _modes_, e.g. the `'authorsFull'` tag. Then if multiple are returned an edge is created between them. So in the case of the author tag `'authorsFull'` a co-authorship network is created. Then for each other tag the entries are also added and edges between the first tag's node and theirs are created.

        The number of times each object occurs is count if _nodeCount_ is `True` and the edges count the number of co-occurrences if _edgeWeight_ is `True`. Both are`True` by default.

        **Note** Do not use this for the construction of co-citation networks use [Recordcollection.networkCoCitation()](./classes/RecordCollection.html#metaknowledge.RecordCollection.networkCoCitation) it is more accurate and has more options.

        # Parameters

        _mode_ : `str`

        > A two character WOS tag or one of the full names for a tag

        _nodeCount_ : `optional [bool]`

        > Default `True`, if `True` each node will have an attribute called "count" that contains an int giving the number of time the object occurred.

        _edgeWeight_ : `optional [bool]`

        > Default `True`, if `True` each edge will have an attribute called "weight" that contains an int giving the number of time the two objects co-occurrenced.

        _stemmer_ : `optional [func]`

        > Default `None`, If _stemmer_ is a callable object, basically a function or possibly a class, it will be called for the ID of every node in the graph, all IDs are strings. For example:

        > The function ` f = lambda x: x[0]` if given as the stemmer will cause all IDs to be the first character of their unstemmed IDs. e.g. the title `'Goos-Hanchen and Imbert-Fedorov shifts for leaky guided modes'` will create the node `'G'`.

        # Returns

        `networkx Graph`

        > A networkx Graph with the objects of the tag _mode_ as nodes and their co-occurrences as edges
        """
        stemCheck = False
        if stemmer is not None:
            if isinstance(stemmer, collections.abc.Callable):
                stemCheck = True
            else:
                raise TagError("stemmer must be callable, e.g. a function or class with a __call__ method.")
        count = 0
        progArgs = (0, "Starting to make a {} from {}".format(_networkTypeString, modes))
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            if edgeAttribute is not None:
                grph = nx.MultiGraph()
            else:
                grph = nx.Graph()
            for R in self:
                if PBar:
                    count += 1
                    PBar.updateVal(count / len(self), "Analyzing: " + str(R))
                if edgeAttribute:
                    edgeVals = [str(v) for v in R.get(edgeAttribute, [])]
                if nodeAttribute:
                    nodeVals = [str(v) for v in R.get(nodeAttribute, [])]
                contents = []
                for attr in modes:
                    tmpContents = R.get(attr, [])
                    if isinstance(tmpContents, list):
                        contents += tmpContents
                    else:
                        contents.append(tmpContents)
                if contents is not None:
                    if not isinstance(contents, str) and isinstance(contents, collections.abc.Iterable):
                        if stemCheck:
                            tmplst = [stemmer(str(n)) for n in contents]
                        else:
                            tmplst = [str(n) for n in contents]
                        if len(tmplst) > 1:
                            for i, node1 in enumerate(tmplst):
                                for node2 in tmplst[i + 1:]:
                                    if edgeAttribute:
                                        for edgeVal in edgeVals:
                                            if grph.has_edge(node1, node2, key = edgeVal):
                                                if edgeWeight:
                                                    for i, a in grph.edges[node1, node2].items():
                                                        if a['key'] == edgeVal:
                                                            grph[node1][node2][i]['weight'] += 1
                                                            break
                                            else:
                                                if edgeWeight:
                                                    attrDict = {'key' : edgeVal, 'weight' : 1}
                                                else:
                                                    attrDict = {'key' : edgeVal}
                                                grph.add_edge(node1, node2, **attrDict)
                                    elif edgeWeight:
                                        try:
                                            grph.edges[node1, node2]['weight'] += 1
                                        except KeyError:
                                            grph.add_edge(node1, node2, weight = 1)
                                    else:
                                        if not grph.has_edge(node1, node2):
                                            grph.add_edge(node1, node2)
                                if not grph.has_node(node1):
                                    grph.add_node(node1)
                                if nodeCount:
                                    try:
                                        grph.node[node1]['count'] += 1
                                    except KeyError:
                                        grph.node[node1]['count'] = 1
                                if nodeAttribute:
                                    try:
                                        currentAttrib = grph.node[node1][nodeAttribute]
                                    except KeyError:
                                        grph.node[node1][nodeAttribute] = nodeVals
                                    else:
                                        for nodeValue in (n for n in nodeVals if n not in currentAttrib):
                                            grph.node[node1][nodeAttribute].append(nodeValue)
                        elif len(tmplst) == 1:
                            if nodeCount:
                                try:
                                    grph.node[tmplst[0]]['count'] += 1
                                except KeyError:
                                    grph.add_node(tmplst[0], count = 1)
                            else:
                                if not grph.has_node(tmplst[0]):
                                    grph.add_node(tmplst[0])
                            if nodeAttribute:
                                try:
                                    currentAttrib = grph.node[tmplst[0]][nodeAttribute]
                                except KeyError:
                                    grph.node[tmplst[0]][nodeAttribute] = nodeVals
                                else:
                                    for nodeValue in (n for n in nodeVals if n not in currentAttrib):
                                        grph.node[tmplst[0]][nodeAttribute].append(nodeValue)
                        else:
                            pass
                    else:
                        if stemCheck:
                            nodeVal = stemmer(str(contents))
                        else:
                            nodeVal = str(contents)
                        if nodeCount:
                            try:
                                grph.node[nodeVal]['count'] += 1
                            except KeyError:
                                grph.add_node(nodeVal, count = 1)
                        else:
                            if not grph.has_node(nodeVal):
                                grph.add_node(nodeVal)
                        if nodeAttribute:
                            try:
                                currentAttrib = grph.node[nodeVal][nodeAttribute]
                            except KeyError:
                                grph.node[nodeVal][nodeAttribute] = nodeVals
                            else:
                                for nodeValue in (n for n in nodeVals if n not in currentAttrib):
                                    grph.node[nodeVal][nodeAttribute].append(nodeValue)
            if PBar:
                PBar.finish("Done making a {} from {}".format(_networkTypeString, modes))
        return grph


    def networkOneMode(self, mode, nodeCount = True, edgeWeight = True, stemmer = None, edgeAttribute = None, nodeAttribute = None):
        """Creates a network of the objects found by one tag _mode_. This is the same as [networkMultiLevel()](#metaknowledge.CollectionWithIDs.networkMultiLevel) with only one tag.

        A **networkOneMode**() looks are each entry in the collection and extracts its values for the tag given by _mode_, e.g. the `'authorsFull'` tag. Then if multiple are returned an edge is created between them. So in the case of the author tag `'authorsFull'` a co-authorship network is created.

        The number of times each object occurs is count if _nodeCount_ is `True` and the edges count the number of co-occurrences if _edgeWeight_ is `True`. Both are`True` by default.

        **Note** Do not use this for the construction of co-citation networks use [Recordcollection.networkCoCitation()](./classes/RecordCollection.html#metaknowledge.RecordCollection.networkCoCitation) it is more accurate and has more options.

        # Parameters

        _mode_ : `str`

        > A two character WOS tag or one of the full names for a tag

        _nodeCount_ : `optional [bool]`

        > Default `True`, if `True` each node will have an attribute called "count" that contains an int giving the number of time the object occurred.

        _edgeWeight_ : `optional [bool]`

        > Default `True`, if `True` each edge will have an attribute called "weight" that contains an int giving the number of time the two objects co-occurrenced.

        _stemmer_ : `optional [func]`

        > Default `None`, If _stemmer_ is a callable object, basically a function or possibly a class, it will be called for the ID of every node in the graph, all IDs are strings. For example:

        > The function ` f = lambda x: x[0]` if given as the stemmer will cause all IDs to be the first character of their unstemmed IDs. e.g. the title `'Goos-Hanchen and Imbert-Fedorov shifts for leaky guided modes'` will create the node `'G'`.

        # Returns

        `networkx Graph`

        > A networkx Graph with the objects of the tag _mode_ as nodes and their co-occurrences as edges
        """
        return self.networkMultiLevel(mode, nodeCount = nodeCount, edgeWeight = edgeWeight, stemmer = stemmer, edgeAttribute = edgeAttribute, nodeAttribute = nodeAttribute, _networkTypeString = 'one mode network')

    def networkTwoMode(self, tag1, tag2, directed = False, recordType = True, nodeCount = True, edgeWeight = True, stemmerTag1 = None, stemmerTag2 = None, edgeAttribute = None):
        """Creates a network of the objects found by two WOS tags _tag1_ and _tag2_, each node marked by which tag spawned it making the resultant graph bipartite.

        A **networkTwoMode()** looks at each Record in the `RecordCollection` and extracts its values for the tags given by _tag1_ and _tag2_, e.g. the `'WC'` and `'LA'` tags. Then for each object returned by each tag and edge is created between it and every other object of the other tag. So the WOS defined subject tag `'WC'` and language tag `'LA'`, will give a two-mode network showing the connections between subjects and languages. Each node will have an attribute call `'type'` that gives the tag that created it or both if both created it, e.g. the node `'English'` would have the type attribute be `'LA'`.

        The number of times each object occurs is count if _nodeCount_ is `True` and the edges count the number of co-occurrences if _edgeWeight_ is `True`. Both are`True` by default.

        The _directed_ parameter if `True` will cause the network to be directed with the first tag as the source and the second as the destination.

        # Parameters

        _tag1_ : `str`

        > A two character WOS tag or one of the full names for a tag, the source of edges on the graph

        _tag1_ : `str`

        > A two character WOS tag or one of the full names for a tag, the target of edges on the graph

        _directed_ : `optional [bool]`

        > Default `False`, if `True` the returned network is directed

        _nodeCount_ : `optional [bool]`

        > Default `True`, if `True` each node will have an attribute called "count" that contains an int giving the number of time the object occurred.

        _edgeWeight_ : `optional [bool]`

        > Default `True`, if `True` each edge will have an attribute called "weight" that contains an int giving the number of time the two objects co-occurrenced.

        _stemmerTag1_ : `optional [func]`

        > Default `None`, If _stemmerTag1_ is a callable object, basically a function or possibly a class, it will be called for the ID of every node given by _tag1_ in the graph, all IDs are strings.

        > For example: the function `f = lambda x: x[0]` if given as the stemmer will cause all IDs to be the first character of their unstemmed IDs. e.g. the title `'Goos-Hanchen and Imbert-Fedorov shifts for leaky guided modes'` will create the node `'G'`.

        _stemmerTag2_ : `optional [func]`

        > Default `None`, see _stemmerTag1_ as it is the same but for _tag2_

        # Returns

        `networkx Graph or networkx DiGraph`

        > A networkx Graph with the objects of the tags _tag1_ and _tag2_ as nodes and their co-occurrences as edges.
        """
        if not isinstance(tag1, str):
            raise TagError("{} is not a string it cannot be a tag.".format(tag1))
        if not isinstance(tag2, str):
            raise TagError("{} is not a string it cannot be a tag.".format(tag2))
        if stemmerTag1 is not None:
            if isinstance(stemmerTag1, collections.abc.Callable):
                stemCheck = True
            else:
                raise TagError("stemmerTag1 must be callable, e.g. a function or class with a __call__ method.")
        else:
            stemmerTag1 = lambda x: x
        if stemmerTag2 is not None:
            if isinstance(stemmerTag2, collections.abc.Callable):
                stemCheck = True
            else:
                raise TagError("stemmerTag2 must be callable, e.g. a function or class with a __call__ method.")
        else:
            stemmerTag2 = lambda x: x
        count = 0
        progArgs = (0, "Starting to make a two mode network of " + tag1 + " and " + tag2)
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            if edgeAttribute is not None:
                if directed:
                    grph = nx.MultiDiGraph()
                else:
                    grph = nx.MultiGraph()
            else:
                if directed:
                    grph = nx.DiGraph()
                else:
                    grph = nx.Graph()
            for R in self:
                if PBar:
                    count += 1
                    PBar.updateVal(count / len(self), "Analyzing: {}".format(R))
                if edgeAttribute is not None:
                    edgeVals = R.get(edgeAttribute, [])
                    if not isinstance(edgeVals, list):
                        edgeVals = [edgeVals]
                contents1 = R.get(tag1)
                contents2 = R.get(tag2)
                if isinstance(contents1, list):
                    contents1 = [stemmerTag1(str(v)) for v in contents1]
                elif contents1 == None:
                    contents1 = []
                else:
                    contents1 = [stemmerTag1(str(contents1))]
                if isinstance(contents2, list):
                    contents2 = [stemmerTag2(str(v)) for v in contents2]
                elif contents2 == None:
                    contents2 = []
                else:
                    contents2 = [stemmerTag2(str(contents2))]
                for node1 in contents1:
                    for node2 in contents2:
                        if edgeAttribute:
                            for edgeVal in edgeVals:
                                if grph.has_edge(node1, node2, key = edgeVal):
                                    if edgeWeight:
                                        grph.edges[node1, node2, edgeVal]['weight'] += 1
                                else:
                                    if edgeWeight:
                                        attrDict = {'key' : edgeVal, 'weight' : 1}
                                    else:
                                        attrDict = {'key' : edgeVal}
                                    grph.add_edge(node1, node2, **attrDict)
                        elif edgeWeight:
                            try:
                                grph.edges[node1, node2]['weight'] += 1
                            except KeyError:
                                grph.add_edge(node1, node2, weight = 1)
                        else:
                            if not grph.has_edge(node1, node2):
                                grph.add_edge(node1, node2)
                    if nodeCount:
                        try:
                            grph.node[node1]['count'] += 1
                        except KeyError:
                            try:
                                grph.node[node1]['count'] = 1
                                if recordType:
                                    grph.node[node1]['type'] = tag1
                            except KeyError:
                                if recordType:
                                    grph.add_node(node1, type = tag1)
                                else:
                                    grph.add_node(node1)
                    else:
                        if not grph.has_node(node1):
                            if recordType:
                                grph.add_node(node1, type = tag1)
                            else:
                                grph.add_node(node1)
                        elif recordType:
                            if 'type' not in grph.node[node1]:
                                grph.node[node1]['type'] = tag1

                for node2 in contents2:
                    if nodeCount:
                        try:
                            grph.node[node2]['count'] += 1
                        except KeyError:
                            try:
                                grph.node[node2]['count'] = 1
                                if recordType:
                                    grph.node[node2]['type'] = tag2
                            except KeyError:
                                grph.add_node(node2, count = 1)
                                if recordType:
                                    grph.node[node2]['type'] = tag2
                    else:
                        if not grph.has_node(node2):
                            if recordType:
                                grph.add_node(node2, type = tag2)
                            else:
                                grph.add_node(node2)
                        elif recordType:
                            if 'type' not in grph.node[node2]:
                                grph.node[node2]['type'] = tag2
            if PBar:
                PBar.finish("Done making a two mode network of " + tag1 + " and " + tag2)
        return grph

    def networkMultiMode(self, *tags, recordType = True, nodeCount = True, edgeWeight = True, stemmer = None, edgeAttribute = None):
        """Creates a network of the objects found by all tags in _tags_, each node is marked by which tag spawned it making the resultant graph n-partite.

        A **networkMultiMode()** looks are each item in the collection and extracts its values for the tags given by _tags_. Then for all objects returned an edge is created between them, regardless of their type. Each node will have an attribute call `'type'` that gives the tag that created it or both if both created it, e.g. if `'LA'` were in _tags_ node `'English'` would have the type attribute be `'LA'`.

        For example if _tags_ was set to `['CR', 'UT', 'LA']`, a three mode network would be created, composed of a co-citation network from the `'CR'` tag. Then each citation would also have edges to all the languages of Records that cited it and to the WOS number of the those Records.

        The number of times each object occurs is count if _nodeCount_ is `True` and the edges count the number of co-occurrences if _edgeWeight_ is `True`. Both are`True` by default.

        # Parameters

        _tags_ : `str`, `str`, `str`, ... or `list [str]`

        > Any number of tags, or a list of tags

        _nodeCount_ : `optional [bool]`

        > Default `True`, if `True` each node will have an attribute called `'count'` that contains an int giving the number of time the object occurred.

        _edgeWeight_ : `optional [bool]`

        > Default `True`, if `True` each edge will have an attribute called `'weight'` that contains an int giving the number of time the two objects co-occurrenced.

        _stemmer_ : `optional [func]`

        > Default `None`, If _stemmer_ is a callable object, basically a function or possibly a class, it will be called for the ID of every node in the graph, note that all IDs are strings.

        > For example: the function `f = lambda x: x[0]` if given as the stemmer will cause all IDs to be the first character of their unstemmed IDs. e.g. the title `'Goos-Hanchen and Imbert-Fedorov shifts for leaky guided modes'` will create the node `'G'`.

        # Returns

        `networkx Graph`

        > A networkx Graph with the objects of the tags _tags_ as nodes and their co-occurrences as edges
        """
        if len(tags) == 1:
            if not isinstance(tags[0], str):
                try:
                    tags = list(tags[0])
                except TypeError:
                    raise TagError("'{}' is not a string it cannot be a tag.".format(tags[0]))
        for t in (i for i in tags if not isinstance(i, str)):
            raise TagError("{} is not a string it cannot be a tag.".format(t))
        stemCheck = False
        if stemmer is not None:
            if isinstance(stemmer, collections.abc.Callable):
                stemCheck = True
            else:
                raise TagError("stemmer must be Callable, e.g. a function or class with a __call__ method.")
        count = 0
        progArgs = (0, "Starting to make a " + str(len(tags)) + "-mode network of: " + ', '.join(tags))
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            if edgeAttribute is not None:
                grph = nx.MultiGraph()
            else:
                grph = nx.Graph()
            for R in self:
                if PBar:
                    count += 1
                    PBar.updateVal(count / len(self), "Analyzing: " + str(R))
                contents = []
                for t in tags:
                    tmpVal = R.get(t)
                    if stemCheck:
                        if tmpVal:
                            if isinstance(tmpVal, list):
                                contents.append((t, [stemmer(str(v)) for v in tmpVal]))
                            else:
                                contents.append((t, [stemmer(str(tmpVal))]))
                    else:
                        if tmpVal:
                            if isinstance(tmpVal, list):
                                contents.append((t, [str(v) for v in tmpVal]))
                            else:
                                contents.append((t, [str(tmpVal)]))
                for i, vlst1 in enumerate(contents):
                    for node1 in vlst1[1]:
                        for vlst2 in contents[i + 1:]:
                            for node2 in vlst2[1]:
                                if edgeAttribute:
                                    for edgeVal in edgeVals:
                                        if grph.has_edge(node1, node2, key = edgeVal):
                                            if edgeWeight:
                                                for i, a in grph.edges[node1, node2].items():
                                                    if a['key'] == edgeVal:
                                                        grph[node1][node2][i]['weight'] += 1
                                                        break
                                        else:
                                            if edgeWeight:
                                                attrDict = {'key' : edgeVal, 'weight' : 1}
                                            else:
                                                attrDict = {'key' : edgeVal}
                                            grph.add_edge(node1, node2, **attrDict)
                                elif edgeWeight:
                                    try:
                                        grph.edges[node1, node2]['weight'] += 1
                                    except KeyError:
                                        grph.add_edge(node1, node2, weight = 1)
                                else:
                                    if not grph.has_edge(node1, node2):
                                        grph.add_edge(node1, node2)
                        if nodeCount:
                            try:
                                grph.node[node1]['count'] += 1
                            except KeyError:
                                try:
                                    grph.node[node1]['count'] = 1
                                    if recordType:
                                        grph.node[node1]['type'] = vlst1[0]
                                except KeyError:
                                    if recordType:
                                        grph.add_node(node1, type = vlst1[0])
                                    else:
                                        grph.add_node(node1)
                        else:
                            if not grph.has_node(node1):
                                if recordType:
                                    grph.add_node(node1, type = vlst1[0])
                                else:
                                    grph.add_node(node1)
                            elif recordType:
                                try:
                                    grph.node[node1]['type'] += vlst1[0]
                                except KeyError:
                                    grph.node[node1]['type'] = vlst1[0]
            if PBar:
                PBar.finish("Done making a {}-mode network of: {}".format(len(tags), ', '.join(tags)))
        return grph



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
"""This module handles the abbreviations, known as J29 abbreviations and given by the J9 tag in WOS Records and for journal titles that WOS employs in citations.

The citations provided by WOS used abbreviated journal titles instead of the full names. The full list of abbreviations can be found at a series pages divided by letter starting at [images.webofknowledge.com/WOK46/help/WOS/A_abrvjt.html](http://images.webofknowledge.com/WOK46/help/WOS/A_abrvjt.html). The function [updatej9DB()](#metaknowledge.journalAbbreviations.backend.getj9dict) is used to scape and parse the pages, it must be run without error before the other features can be used. _metaknowledge_. If the database is requested by `getj9dict()`, which is what [Citations](../classes/Citation.html#metaknowledge.citation.Citation) use, and the database is not found or is corrupted then [updatej9DB()](#metaknowledge.journalAbbreviations.backend.updatej9DB) will be run to download the database if this fails an `mkException` will be raised, the download and parsing usually takes less than a second on a good internet connection.

The other functions of the module are for manually adding and removing abbreviations from the database. It is recommended that this be done with the command-line tool `metaknowledge` instead of with a script.
"""

from .backend import getj9dict, abrevDBname, manualDBname, addToDB



import os
import sys
import time
import math
import threading

class _ProgressBar(object):
    difTermAndBar = 8 #the number of characters difference between the bar's length and the terminal's width
    timeLength = 6 # width of elapse time display
    percLength = 6 # width of percent display
    def __init__(self, initPer, initString = ' ', output = sys.stdout, secondRow = False, dummy = False):
        self.dummy = dummy
        self.finished = False
        self.big = secondRow
        self.per = initPer
        self.out = output
        self.inputString = initString
        if not dummy:
            self.sTime = time.time()
            try:
                self.barMaxLength = os.get_terminal_size(self.out.fileno()).columns - self.difTermAndBar
                if self.barMaxLength < 0:
                    self.barMaxLength = 0
            except OSError:
                self.barMaxLength = 80 - self.difTermAndBar
            except AttributeError:
                #Pypy fallback
                self.barMaxLength = 80 - self.difTermAndBar
            self.ioThread = threading.Thread(target = self.threadedUpdate, kwargs = {"self" : self})
            self.ioThread.daemon = True
            self.ioThread.start()

    def __bool__(self):
        return not self.dummy

    def __del__(self):
        if not self.dummy and not self.finished:
            self.finished = True
            self.ioThread.join()
            self.out.write('\n')
            self.out.flush()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__del__()

    def updateVal(self, inputPer, inputString = None):
        self.per = inputPer
        if inputString is not None:
            self.inputString = inputString

    def finish(self, inputString):
        if not self.dummy:
            self.finished = True
            self.inputString = str(inputString)
            self.ioThread.join()
            try:
                self.barMaxLength = os.get_terminal_size(self.out.fileno()).columns - self.difTermAndBar
                if self.barMaxLength < 0:
                    self.barMaxLength = 0
            except OSError:
                self.barMaxLength = 80 - self.difTermAndBar
            except AttributeError:
                #Pypy fallback
                self.barMaxLength = 80 - self.difTermAndBar
            if self.big:
                self.out.write('\n' + ' ' * (self.barMaxLength + self.difTermAndBar) + '\033[F')
            else:
                self.out.write('\r')
            if len(self.inputString) < self.barMaxLength + self.difTermAndBar - self.timeLength:
                tString = self.prepTime(time.time() - self.sTime, self.barMaxLength + self.difTermAndBar - len(self.inputString) - 1)
                self.out.write(self.inputString + ' ' + tString)
            else:
                self.out.write(self.prepString(self.inputString, self.barMaxLength + self.difTermAndBar - self.timeLength) + self.prepTime(time.time() - self.sTime, self.timeLength))
            self.out.write('\n')
            self.out.flush()

    def jumpUp(self):
        self.out.write('\033[F')
        self.out.flush()

    @staticmethod
    def prepString(s, maxLength):
        maxLength = maxLength - 1
        sString = str(s)
        if len(sString) <= maxLength:
            return sString.ljust(maxLength, ' ') + ' '
        else:
            if maxLength % 2 == 0:
                return sString[:int(maxLength/2 - 3)] + '...' + sString[int(-maxLength/2):] + ' '
            else:
                return sString[:int(maxLength/2 - 2)] + '...' + sString[int(-maxLength/2):] + ' '

    @staticmethod
    def prepTime(t, maxLength):
        try:
            if math.log10(t) + 3.01 > maxLength:
                return "{1:{0}.0E}s".format(maxLength - 1 ,t)
            else:
                return "{1:{0}.1f}s".format(maxLength - 1,t)
        except ValueError:
            return "{1:{0}.1f}s".format(maxLength - 1,t)

    @staticmethod
    def threadedUpdate(self = None):
        while not self.finished:
            try:
                self.barMaxLength = os.get_terminal_size(self.out.fileno()).columns - self.difTermAndBar
                if self.barMaxLength < 0:
                    self.barMaxLength = 0
            except OSError:
                self.barMaxLength = 80 - self.difTermAndBar
            except AttributeError:
                #Pypy fallback
                self.barMaxLength = 80 - self.difTermAndBar
            self.out.write('\r')
            percentString = '{:.1%}'.format(self.per).rjust(self.percLength, ' ')
            barLength = int(self.per * self.barMaxLength)
            if self.big and self.inputString:
                self.dString = self.prepString(self.inputString, self.barMaxLength + self.difTermAndBar - self.timeLength) + self.prepTime(time.time() - self.sTime, self.timeLength)
                if barLength >= self.barMaxLength:
                    self.out.write('[' + '=' * barLength + ']' + percentString)
                    self.out.write('\n' + self.dString + '\033[F')
                else:
                    self.out.write('[' + '=' * barLength + '>' + ' ' * (self.barMaxLength - barLength - 1) + ']' + percentString)
                    self.out.write('\n' + self.dString + '\033[')
            elif self.inputString:
                self.dString = self.prepString(self.inputString, self.barMaxLength + self.difTermAndBar - self.timeLength - self.percLength - 2) + '[' + self.prepTime(time.time() - self.sTime, self.timeLength) +  ']' + percentString
                self.out.write(self.dString)
            else:
                if barLength >= self.barMaxLength:
                    self.out.write('[' + '=' * barLength + ']' + percentString + '\r')
                else:
                    self.out.write('[' + '=' * barLength + '>' + ' ' * (self.barMaxLength - barLength - 1) + ']' + percentString + '\r')
            self.out.flush()
            time.sleep(.1)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import unittest
import metaknowledge


class TestDiffusion(unittest.TestCase):
    def setUp(self):
        metaknowledge.VERBOSE_MODE = False
        self.RC = metaknowledge.RecordCollection("metaknowledge/tests/testFile.isi")

    def test_diffusionGraph(self):
        G = metaknowledge.diffusionGraph(self.RC, self.RC)
        Gcr_ut = metaknowledge.diffusionGraph(self.RC, self.RC, sourceType = "CR", targetType = "UT")
        self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 42 nodes, 1569 edges, 0 isolates, 35 self loops, a density of 0.91115 and a transitivity of 0.894934')
        self.assertEqual(metaknowledge.graphStats(Gcr_ut, sentenceString = True), 'The graph has 528 nodes, 3591 edges, 246 isolates, 0 self loops, a density of 0.0129054 and a transitivity of 0')

    def test_multiGraph(self):
        G = metaknowledge.diffusionGraph(self.RC, self.RC, labelEdgesBy = 'PY')
        metaknowledge.dropEdges(G, dropSelfLoops = True)
        #multigraphs have issues their edge counts are somewhat unpredictable
        self.assertEqual(metaknowledge.graphStats(G, stats = ('nodes', 'isolates', 'loops'), sentenceString = True), 'The graph has 42 nodes, 0 isolates and 0 self loops')

    def test_diffusionCounts(self):
        d = metaknowledge.diffusionCount(self.RC, self.RC)
        dc = metaknowledge.diffusionCount(self.RC, self.RC, compareCounts = True)
        dWC = metaknowledge.diffusionCount(self.RC, self.RC, sourceType = "WC")
        self.assertIsInstance(d.keys().__iter__().__next__(), metaknowledge.Record)
        self.assertTrue(-1 < d.values().__iter__().__next__() < 10)
        self.assertIsInstance(list(dWC.keys())[0], str)
        self.assertTrue(-1 < dWC.values().__iter__().__next__() < 24)
        for t in dc.values():
            self.assertEqual(t[0], t[1])

    def test_diffusionPandas(self):
        d = metaknowledge.diffusionCount(self.RC, self.RC, pandasFriendly = True)
        dwc = metaknowledge.diffusionCount(self.RC, self.RC, pandasFriendly = True, sourceType = "WC", compareCounts = True)
        dyear = metaknowledge.diffusionCount(self.RC, self.RC, pandasFriendly = True, extraValue = 'year')
        self.assertTrue("TI" in d.keys())
        self.assertEqual(len(d), 44)
        self.assertTrue(len(d["UT"]), len(self.RC))
        self.assertTrue("WC" in dwc)
        self.assertEqual(3, len(dwc))
        self.assertEqual(len(dwc["TargetCount"]), 9)
        self.assertEqual(dwc["TargetCount"], dwc["SourceCount"])
        self.assertEqual(len(dyear), len(d) + 1)
        self.assertNotEqual(dyear["TargetCount"], dwc["SourceCount"])
        self.assertEqual(len([c for c in dyear["TargetCount"] if c > 1]), 9)
        self.assertTrue(1979 in dyear['year'])



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
import zipfile
import io
import csv
import os.path
import urllib

from ..mkExceptions import GenderException

dataURL = 'https://github.com/UWNETLAB/globalnamedata/archive/0.3.zip'
americanNamesPath = 'globalnamedata-0.3/assets/usprocessed.csv'
ukNamesPath = 'globalnamedata-0.3/assets/ukprocessed.csv'

targetFilePath = os.path.join(os.path.normpath(os.path.dirname(__file__)), 'namesData.csv')

csvFields = [
    'Name',
    'years.appearing',
    'count.male',
    'count.female',
    'prob.gender',
    'obs.male',
    'est.male',
    'upper',
    'lower'
]

#global to reduce need to reload dict
mappingDict = None

def downloadData(useUK = False):
    zipFile = io.BytesIO(urllib.request.urlopen(dataURL).read())
    if useUK:
        namesFile = zipfile.ZipFile(zipFile).open(ukNamesPath)
    else:
        namesFile = zipfile.ZipFile(zipFile).open(americanNamesPath)
    try:
        with open(targetFilePath, 'wb') as f:
            f.write(namesFile.read())
    except PermissionError:
        raise PermissionError("Can not write to {}, you try rerunning with higher privileges".format(targetFilePath))

def getMapping(useUK = False):
    if not os.path.isfile(targetFilePath):
        downloadData(useUK)
    retDict = {}
    with open(targetFilePath) as f:
        reader = csv.DictReader(f, fieldnames = csvFields)
        next(reader)
        for line in reader:
            retDict[line['Name'].title()] = line['prob.gender']
    return retDict

def nameStringGender(s, noExcept = False):
    """Expects `first, last`"""
    global mappingDict
    try:
        first = s.split(', ')[1].split(' ')[0].title()
    except IndexError:
        if noExcept:
            return 'Unknown'
        else:
            return GenderException("The given String: '{}' does not have a  last name, first name pair in with a ', ' seperation.".format(s))
    if mappingDict is None:
        mappingDict = getMapping()
    return mappingDict.get(first, 'Unknown')

def recordGenders(R):
    return {auth : nameStringGender(auth, noExcept = True) for auth in R.get('authorsFull', [])}



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import unittest
import metaknowledge

class TestRecord(unittest.TestCase):
    def setUp(self):
        metaknowledge.VERBOSE_MODE = False
        self.R = metaknowledge.WOSRecord(simplePaperString)
        self.Rbad = metaknowledge.WOSRecord(simplePaperString[:-3])

    def test_isRecord(self):
        self.assertTrue(isinstance(self.R, metaknowledge.WOSRecord))

    def test_base(self):
        metaknowledge.ExtendedRecord.__abstractmethods__ = frozenset()
        R = metaknowledge.ExtendedRecord(self.R._fieldDict, self.R.id, self.R.bad, self.R.error)
        self.assertEqual(R.encoding(), 'utf-8')
        with self.assertRaises(KeyError):
            R.specialFuncs('TI')
        self.assertEqual(R.writeRecord('IF_YOU_SEE_THIS_A_TEST_HAS_GONE_VERY_WRONG_PLEASE_TELL_SOMEONE.mk_test_file_that_you_should_never_see'), None)
        self.assertEqual(R.getAltName('TI'), None)
        self.assertEqual(R.tagProcessingFunc('TI')('A'), 'A')

    def test_bad(self):
        self.assertTrue(self.Rbad.bad)
        with self.assertRaises(TypeError):
            metaknowledge.WOSRecord(set('a','b'))

    def test_equality(self):
        self.assertEqual(self.R, self.R)
        self.assertTrue(self.R != self.Rbad)
        self.assertFalse(self.R == 1)

    def test_contains(self):
        self.assertFalse('WOS' in self.R)
        self.assertTrue('C1' in self.R)

    def test_getitem(self):
        self.assertEqual(self.R['UT'], self.R.id)
        with self.assertRaises(TypeError):
            ret = self.R[1]

    def test_get(self):
        self.assertEqual(self.R.get('VL'), '1')
        self.assertEqual(self.R.get('volume'), '1')
        self.assertEqual(self.R.get('VL', raw = True), ['1'])
        self.assertEqual(self.R.get('volume', raw = True), ['1'])
        self.assertEqual(self.R.get('!@#$%^&*&%$#'), None)

    def test_values(self):
        self.assertTrue('1' in self.R.values())
        self.assertFalse('1' in self.R.values(raw = True))

    def test_items(self):
        self.assertTrue(('volume', '1') in self.R.items())
        self.assertTrue(('VL', ['1']) in self.R.items(raw = True))

    def test_len(self):
        self.assertEqual(len(self.R), 22)

    def test_hash(self):
        self.assertNotEqual(hash(self.R), hash(self.Rbad))

    def test_repr(self):
        self.assertEqual(repr(self.R), "<metaknowledge.WOSRecord object {}>".format(self.R.id))
        self.R.bad = True
        self.assertEqual(repr(self.R), "<metaknowledge.WOSRecord object BAD>")

    def test_binary(self):
        self.assertEqual(bytes(self.R), bytes(simplePaperString, encoding = 'utf-8'))
        self.R.bad = True
        with self.assertRaises(metaknowledge.BadRecord):
            b = bytes(self.R)

    def test_state(self):
        state = self.R.__getstate__()
        Rtmp = metaknowledge.WOSRecord('PT J')
        Rtmp.__setstate__(state)
        self.assertEqual(self.R, Rtmp)

    def test_title(self):
        self.assertEqual(self.R.title, "Example Paper")
        self.assertEqual(str(self.R), "WOSRecord(Example Paper)")

    def test_author(self):
        R2 = self.R.copy()
        del R2._fieldDict['AF']
        self.assertEqual(self.R['authorsFull'], ["John, Doe"])
        self.assertEqual(self.R.authors, ["John, Doe"])
        self.assertEqual(R2.authors, ['John, D'])

    def test_year(self):
        self.assertEqual(self.R['year'], 2015)

    def test_month(self):
        self.assertEqual(self.R['month'], 4)

    def test_cites(self):
        self.assertEqual(str(self.R['citations'][0]), "John D. 1999, TOPICS IN COGNITIVE SCIENCE")

        self.assertEqual(len(self.R['citations']), 1)

    def test_WOS(self):
        self.assertEqual(self.R['wosString'], 'WOS:123317623000007')

    def test_citationGen(self):
        R2 = self.R.copy()
        R2._fieldDict['AU'] = ['John D', 'John H']
        R0 = self.R.copy()
        del R0._fieldDict['AU']
        del R0._fieldDict['AF']
        del R0._fieldDict['J9']
        self.assertTrue(self.R.createCitation() == metaknowledge.Citation("John D, 2015, EXAMPLE, V1, P1, DOI 10.1111"))
        self.assertTrue(R0.createCitation(multiCite = True)[0] == metaknowledge.Citation("2015, Example Paper, V1, P1, DOI 10.1111"))
        self.assertTrue(set(R2.createCitation(multiCite = True)) == set((self.R.createCitation(), metaknowledge.Citation("John H, 2015, EXAMPLE, V1, P1, DOI 10.1111"))))

    def test_journal(self):
        self.assertEqual(self.R['journal'], 'TOPICS IN COGNITIVE SCIENCE')

    def test_tags(self):
        self.assertEqual(self.R.subDict(['a', 'b']), {'a': None, 'b': None})

    def test_specials(self):
        for t in metaknowledge.commonRecordFields:
            if t not in ['id', 'address', 'grants', 'selfCitation']:
                self.assertIsInstance(self.R.getAltName(t), str)

simplePaperString = """PT J
AU John, D
AF John, Doe
TI Example Paper
SO TOPICS IN COGNITIVE SCIENCE
LA English
DT Article
DE Example; testing
ID REAL; TIME
AB This is a test.
C1 UW, Ontario, Canada.
RP John, D (reprint author), UW, Ontario, Canada.
CR John D. 1999, TOPICS IN COGNITIVE SCIENCE
J9 EXAMPLE
JI examaple
PD APR
PY 2015
VL 1
BP 1
EP 2
DI 10.1111
UT WOS:123317623000007
ER
"""



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import unittest
import metaknowledge
import os
import io
import sys
from metaknowledge.progressBar import _ProgressBar

fileShortName = 'testNetworks'
fileEName = 'testNetworks_edgeList.tst'
fileNName = 'testNetworks_nodeAttributes.tst'
filesuffix = 'tst'

class TestHelpers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.RCmain = metaknowledge.RecordCollection("metaknowledge/tests/testFile.isi")
        cls.Gmain = cls.RCmain.networkCoCitation()

    def setUp(self):
        metaknowledge.VERBOSE_MODE = False
        self.RC = self.RCmain.copy()
        self.G = self.Gmain.copy()

    def test_graphwrite(self):
        metaknowledge.writeGraph(self.G, fileShortName, suffix = filesuffix)
        tmpG = metaknowledge.readGraph(fileEName, fileNName)
        self.assertEqual(len(tmpG.edges()), len(self.G.edges()))
        self.assertEqual(len(tmpG.nodes()), len(self.G.nodes()))
        os.remove(fileEName)
        os.remove(fileNName)

    def test_tnetWriter(self):
        fName = fileShortName + "_tnet.csv"
        G = self.RC.networkTwoMode('AF', 'WC', edgeAttribute = 'PY')
        metaknowledge.writeTnetFile(G, fName, 'type', weighted = True, timeString = 'key')
        self.assertAlmostEqual(os.path.getsize(fName), 1015, delta=100)
        os.remove(fName)
        metaknowledge.writeTnetFile(G, fName, 'type')
        self.assertAlmostEqual(os.path.getsize(fName), 378, delta=50)
        os.remove(fName)

    def test_progress(self):
        metaknowledge.VERBOSE_MODE = True
        tmpIO = io.StringIO()
        P = _ProgressBar(0, "testing", output = tmpIO, dummy = True)
        metaknowledge.writeEdgeList(self.G, fileEName, _progBar = P, )
        tmpIO.seek(0)
        s = ''.join(tmpIO.readlines())
        self.assertEqual(len(s), 0)
        P = _ProgressBar(0, "testing", output = tmpIO)
        metaknowledge.writeEdgeList(self.G, fileEName, _progBar = P)
        tmpIO.seek(0)
        os.remove(fileEName)
        s = ''.join(tmpIO.readlines())
        self.assertEqual(s[-14], '[')
        self.assertEqual(s[-1], '%')
        P.finish("done test")
        tmpIO.seek(0)
        s = ''.join(tmpIO.readlines())
        self.assertEqual(s[-81:-3], 'done test                                                                   0.')
        metaknowledge.VERBOSE_MODE = False

    def test_dropEdges(self):
        metaknowledge.dropEdges(self.G, minWeight = 1, maxWeight = 3, dropSelfLoops = True)
        self.assertEqual(metaknowledge.graphStats(self.G, sentenceString = True), "The graph has 493 nodes, 12711 edges, 0 isolates, 0 self loops, a density of 0.104809 and a transitivity of 0.588968")
        self.assertTrue(self.G.edges['Imbert C, 1975, NOUV REV OPT', 'Fainman Y, 1984, APPL OPTICS']['weight'] == 1)

    def test_dropNodeByCount(self):
        metaknowledge.dropNodesByCount(self.G, minCount = 2, maxCount = 5)
        self.assertEqual(metaknowledge.graphStats(self.G, sentenceString = True), "The graph has 106 nodes, 1205 edges, 0 isolates, 17 self loops, a density of 0.218149 and a transitivity of 0.751036")
        self.assertTrue(self.G.node['Shih H, 1971, PHYS REV A']['count'] == 2)

    def test_dropNodesByDegree(self):
        metaknowledge.dropNodesByDegree(self.G, minDegree = 20, maxDegree = 100)
        self.assertEqual(metaknowledge.graphStats(self.G, sentenceString = True), "The graph has 385 nodes, 5923 edges, 0 isolates, 11 self loops, a density of 0.0802083 and a transitivity of 0.954487")
        self.assertTrue(self.G.edges['Mazur P, 1953, MEM ACAD ROY BELG', 'Livens Gh, 1948, P CAMB PHILOS SOC']['weight'] == 1)

    def test_mergeGraphs(self):
        RC1 = self.RC.yearSplit(0,1978)
        RC2 = self.RC.yearSplit(1979,10000)
        G1 = RC1.networkCoCitation()
        G2 = RC2.networkCoCitation()
        metaknowledge.mergeGraphs(G1,G2)
        for node, attr in G1.nodes(data = True):
            self.assertEqual(self.G.node[node]['count'], attr['count'])
        for node1, node2, attr in G1.edges(data = True):
            self.assertEqual(self.G.edges[node1, node2]['weight'], attr['weight'])



from ...WOS.tagProcessing.helpFuncs import makeBiDirectional


authorBasedTags = [
    'AD',
    'AUID',
]

tagNameDict = {
"Abstract" : "AB",
"CopyrightInformation" : "CI",
"Affiliation" : "AD",
"InvestigatorAffiliation" : "IRAD",
"ArticleIdentifier" : "AID",
"Author" : "AU",
"AuthorIdentifier" : "AUID",
"FullAuthor" : "FAU",
"BookTitle" : "BTI",
"CollectionTitle" : "CTI",
"CorporateAuthor" : "CN",
"CreateDate" : "CRDT",
"DateCompleted" : "DCOM",
"DateCreated" : "DA",
"DateLastRevised" : "LR",
"DateElectronicPublication" : "DEP",
"DatePublication" : "DP",
"Edition" : "EN",
"Editor" : "ED",
"Editor" : "FED",
"EntrezDate" : "EDAT",
"GeneSymbol" : "GS",
"GeneralNote" : "GN",
"GrantNumber" : "GR",
"Investigator" : "IR",
"InvestigatorFull" : "FIR",
"ISBN" : "ISBN",
"ISSN" : "IS",
"Issue" : "IP",
"JournalTitleAbbreviation" : "TA",
"JournalTitle" : "JT",
"Language" : "LA",
"LocationIdentifier" : "LID",
"ManuscriptIdentifier" : "MID",
"MeSHDate" : "MHDA",
"MeSHTerms" : "MH",
"NLMID" : "JID",
"NumberReferences" : "RF",
"OtherAbstract" : "OAB",
"OtherAbstract" : "OABL",
"OtherCopyright" : "OCI",
"OtherID" : "OID",
"OtherTerm" : "OT",
"OtherTermOwner" : "OTO",
"Owner" : "OWN",
"Pagination" : "PG",
"PersonalNameSubject" : "PS",
"FullPersonalNameSubject" : "FPS",
"PlacePublication" : "PL",
"PublicationHistoryStatus" : "PHST",
"PublicationStatus" : "PST",
"PublicationType" : "PT",
"PublishingModel" : "PUBM",
"PubMedCentralIdentifier" : "PMC",
"PubMedCentralRelease" : "PMCR",
"PubMedUniqueIdentifier" : "PMID",
"RegistryNumber" : "RN",
"SubstanceName" : "NM",
"SecondarySourceID" : "SI",
"Source" : "SO",
"SpaceFlightMission" : "SFM",
"Status" : "STAT",
"Subset" : "SB",
"Title" : "TI",
"TransliteratedTitle" : "TT",
"Volume" : "VI",
"VolumeTitle" : "VTI",
"CommentIn" : "CIN",
"ErratumIn" : "EIN",
"ErratumFor" : "EFR",
"CorrectedRepublishedIn" : "CRI",
"CorrectedRepublishedFrom" : "CRF",
"DatasetIn" : "DDIN",
"DatasetUseReportedIn" : "DRIN",
"PartialRetractionIn" : "PRIN",
"PartialRetractionOf" : "PROF",
"RepublishedIn" : "RPI",
"RepublishedFrom" : "RPF",
"RetractionIn" : "RIN",
"RetractionOf" : "ROF",
"UpdateIn" : "UIN",
"UpdateOf" : "UOF",
"SummaryForPatients" : "SPIN",
"OriginalReportIn" : "ORI",
}

tagNameConverterDict = makeBiDirectional(tagNameDict)



import itertools

from ..mkExceptions import BadPubmedFile

from .recordMedline import MedlineRecord

def isMedlineFile(infile, checkedLines = 2):
    """Determines if _infile_ is the path to a Medline file. A file is considerd to be a Medline file if it has the correct encoding (`latin-1`) and within the first _checkedLines_ a line starts with `"PMID- "`.

    # Parameters

    _infile_ : `str`

    > The path to the targets file

    _checkedLines_ : `optional [int]`

    > default 2, the number of lines to check for the header

    # Returns

    `bool`

    > `True` if the file is a Medline file
    """
    try:
        with open(infile, 'r', encoding='latin-1') as openfile:
            f = enumerate(openfile, start = 0)
            for i in range(checkedLines):
                if f.__next__()[1].startswith("PMID- "):
                    #Only indicator I could find
                    return True
    except (StopIteration, UnicodeDecodeError):
        return False
    else:
        return False

def medlineParser(pubFile):
    """Parses a medline file, _pubFile_, to extract the individual entries as [MedlineRecords](#metaknowledge.medline.recordMedline.MedlineRecord).

    A medline file is a series of entries, each entry is a series of tags. A tag is a 2 to 4 character string each tag is padded with spaces on the left to make it 4 characters which is followed by a dash and a space (`'- '`). Everything after the tag and on all lines after it not starting with a tag is considered associated with the tag. Each entry's first tag is `PMID`, so a first line looks something like `PMID- 26524502`. Entries end with a single blank line.

    # Parameters

    _pubFile_ : `str`

    > A path to a valid medline file, use [isMedlineFile](#metaknowledge.medline.medlineHandlers.isMedlineFile) to verify

    # Returns

    `set[MedlineRecord]`

    > Records for each of the entries
    """
    #assumes the file is MEDLINE
    recSet = set()
    error = None
    lineNum = 0
    try:
        with open(pubFile, 'r', encoding = 'latin-1') as openfile:
            f = enumerate(openfile, start = 1)
            lineNum, line = next(f)
            try:
                while True:
                    if line.startswith("PMID- "):
                        try:
                            r = MedlineRecord(itertools.chain([(lineNum, line)], f), sFile = pubFile, sLine = lineNum)
                            recSet.add(r)
                        except BadPubmedFile as e:
                            badLine = lineNum
                            try:
                                lineNum, line = next(f)
                                while not line.startswith("PMID- "):
                                    lineNum, line = next(f)
                            except (StopIteration, UnicodeDecodeError) as e:
                                if error is None:
                                    error = BadPubmedFile("The file '{}' becomes unparsable after line: {}, due to the error: {} ".format(pubFile, badLine, e))
                                raise e
                    elif line != '\n':
                        if error is None:
                            error = BadPubmedFile("The file '{}' has parts of it that are unparsable starting at line: {}.".format(pubFile, lineNum))
                    lineNum, line = next(f)
            except StopIteration:
                #End of the file has been reached
                pass
    except UnicodeDecodeError:
        if error is None:
            error = BadPubmedFile("The file '{}' has parts of it that are unparsable starting at line: {}.".format(pubFile, lineNum))
    return recSet, error



import os.path
try:
    import collections.abc
except ImportError:
    import collections
    collections.abc = collections

import networkx as nx

from .progressBar import _ProgressBar

from .mkCollection import CollectionWithIDs
from .mkExceptions import GrantCollectionException, BadInputFile, UnknownFile

from .grants.baseGrant import Grant

from .fileHandlers import grantProcessors

import metaknowledge

class GrantCollection(CollectionWithIDs):
    def __init__(self, inGrants = None, name = '', extension = '', cached = False, quietStart = False):

        progArgs = (0, "Starting to make a GrantCollection")
        if metaknowledge.VERBOSE_MODE and not quietStart:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            bad = False
            errors = {}
            name = name
            grantTypes = set()

            if inGrants is None:
                PBar.updateVal(.5, "Empty GrantCollection created")
                if not name:
                    name = "Empty"
                grantsSet = set()

            elif isinstance(inGrants, str):
                if os.path.isfile(inGrants):
                    PBar.updateVal(.2, "GrantCollection from a file started")
                    if not inGrants.endswith(extension):
                        raise GrantCollectionException("extension of input file does not match requested extension '{}'.".format(extension))
                    if not name:
                        name = os.path.splitext(os.path.split(inGrants)[1])[0]
                    try:
                        for grantType, processor, detector in grantProcessors:
                            if detector(inGrants):
                                grantTypes.add(grantType)
                                grantsSet, gError = processor(inGrants)
                                if gError is not None:
                                    bad = True
                                    errors[inGrants] = gError
                                break
                    except UnknownFile:
                        raise BadInputFile("'{}' does not match any known grant file type and the default parser could not handle it.\nIts header might be damaged or it could have been modified by another program.".format(inGrants))
                elif os.path.isdir(inGrants):
                    count = 0
                    PBar.updateVal(0, "GrantCollection from files in {}".format(inGrants))
                    if extension and not name:
                        name = "{}-files-from-{}".format(extension, inGrants)
                    elif not name:
                        name = "files-from-{}".format(inGrants)
                    grantsSet = set()
                    flist = []
                    for f in os.listdir(inGrants):
                        fullF = os.path.join(os.path.abspath(inGrants), f)
                        if fullF.endswith(extension) and os.path.isfile(fullF):
                            flist.append(fullF)
                    if cached:
                        cacheName = os.path.join(inGrants, '{}.[{}].mkGrantDirCache'.format(os.path.basename(os.path.abspath(inGrants)), extension))
                        if self._loadFromCache(cacheName, flist, name, extension):
                            try:
                                PBar.finish("Done reloading {} Grants from cache".format(len(self)))
                            except AttributeError:
                                PBar.finish("Done reloading from the cache {}. Warning an error occured.".format(cacheName))
                            return
                        else:
                            PBar.updateVal(0, 'Cache error, rereading files')
                    for fileName in flist:
                        count += 1
                        PBar.updateVal(count / len(flist), "Reading grants from: {}".format(fileName))
                        try:
                            for grantType, processor, detector in grantProcessors:
                                if detector(fileName):
                                    grantTypes.add(grantType)
                                    grants, gError = processor(fileName)
                                    if gError is not None:
                                        bad = True
                                        errors[fileName] = gError
                                    #This is need for any grants thats can span mutiple files
                                    if grantType == "NSERCGrant":
                                        if len(grants & grantsSet) > 0:
                                            #In theory this could be done with
                                            #the builtin operators (& and ^),
                                            #but the exact results are not
                                            #defined for objects with identical
                                            #hashes without identical attributes
                                            #so it could not be counted on
                                            #Thus we get this mess
                                            overlap = {}
                                            for Gin in grants:
                                                if Gin in grantsSet:
                                                    overlap[Gin.id] = Gin
                                            for Gover in overlap.values():
                                                grants.remove(Gover)
                                            for Gset in grantsSet:
                                                if Gset.id in overlap:
                                                    Gset.update(overlap[Gset.id])
                                    grantsSet |= grants
                                    break
                        except UnknownFile:
                            if extension != '':
                                raise BadInputFile("'{}' does not match any known file type, but has the requested extension '{}'. Its header might be damaged or it could have been modified by another program.".format(fileName, extension))
                            else:
                                pass
                else:
                    raise GrantCollectionException("'{}' is not a path to a directory or file. Strings cannot be used to initialize GrantCollections".format(inGrants))

            elif isinstance(inGrants, collections.abc.Iterable):
                PBar.updateVal(.5, "GrantCollection started from {}".format(type(inGrants).__name__))
                if not name:
                    name = "Grants-from-a-{}".format(type(inGrants).__name__)
                for G in inGrants:
                    if not isinstance(G, Grant):
                        raise GrantCollectionException("A GrantCollection cannot be created from a Iterable containing {}".format(G))
                grantsSet = set(inGrants)
            else:
                raise GrantCollectionException("A GrantCollection cannot be created from {}".format(inGrants))
            CollectionWithIDs.__init__(self, grantsSet, Grant, grantTypes, name, bad, errors, quietStart = quietStart)
            if cached:
                PBar.updateVal(1, "Writing GrantCollection cache to {}".format(cacheName))
                self._createCache(cacheName, flist, name, extension)
            try:
                PBar.finish("Done making a GrantCollection of {} Grants".format(len(self)))
            except AttributeError:
                PBar.finish("Done making a GrantCollection. Warning an error occured.")

    def networkCoInvestigatorInstitution(self, targetTags = None, tagSeperator = ';', count = True, weighted = True):
        """This works the same as [networkCoInvestigator()](#metaknowledge.GrantCollection.networkCoInvestigator) see it for details."""
        return self.networkCoInvestigator(targetTags = targetTags, tagSeperator = tagSeperator, count = count, weighted = weighted, _institutionLevel = True)

    def networkCoInvestigator(self, targetTags = None, tagSeperator = ';', count = True, weighted = True, _institutionLevel = False):
        """Creates a co-investigator from the collection

        Most grants do not have a known investigator tag so it must be provided by the user in _targetTags_ and the separator character if it is not a semicolon should also be given.

        # Parameters

        > _targetTags_ : `optional list[str]`

        > A list of all the Grant tags to check for investigators

        _tagSeperator_ : `optional str`

        > The character that separates the individual investigator's names

        _count_ : `optional bool`

        > Default `True`, if `True` the number of time a name occurs will be given

        _weighted_ : `optional bool`

        > Default `True`, if `True` the edge weights will be calculated and added to the edges

        # Returns

        `networkx Graph`

        > The graph of co-investigator
        """
        grph = nx.Graph()
        pcount = 0
        if _institutionLevel:
            progArgs = (0, "Starting to make a co-institution network")
        else:
            progArgs = (0, "Starting to make a co-investigator network")
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            for G in self:
                if PBar:
                    pcount += 1
                    PBar.updateVal(pcount/ len(self), "Analyzing: " + str(G))
                if _institutionLevel:
                    investList = G.getInstitutions(tags = targetTags, seperator = tagSeperator, _getTag = True)
                else:
                    investList = G.getInvestigators(tags = targetTags, seperator = tagSeperator, _getTag = True)
                if len(investList) > 1:
                    for i, invest1 in enumerate(investList):
                        if invest1[0] not in grph:
                            if count:
                                grph.add_node(invest1[0], count = 1, field = invest1[1])
                            else:
                                grph.add_node(invest1[0], field = invest1[1])
                        elif count:
                            grph.node[invest1[0]]['count'] += 1
                        for invest2 in investList[i + 1:]:
                            if invest2[0] not in grph:
                                if count:
                                    grph.add_node(invest2[0], count = 1, field = invest2[1])
                                else:
                                    grph.add_node(invest2[0], field = invest2[1])
                            elif count:
                                grph.node[invest2[0]]['count'] += 1
                            if grph.has_edge(invest1[0], invest2[0]) and weighted:
                                grph.edges[invest1[0], invest2[0]]['weight'] += 1
                            elif weighted:
                                grph.add_edge(invest1[0], invest2[0], weight = 1)
                            else:
                                grph.add_edge(invest1[0], invest2[0])
                elif len(investList) > 0:
                    invest1 = investList[0]
                    if invest1[0] not in grph:
                        if count:
                            grph.add_node(invest1[0], count = 1, field = invest1[1])
                        else:
                            grph.add_node(invest1[0], field = invest1[1])
                    elif count:
                        grph.node[invest1[0]]['count'] += 1
            if _institutionLevel:
                PBar.finish("Done making a co-institution network from {}".format(self))
            else:
                PBar.finish("Done making a co-investigator network from {}".format(self))
        return grph



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import itertools
import networkx as nx

from .WOS.tagProcessing.funcDicts import tagsAndNameSet, normalizeToTag
from .progressBar import _ProgressBar
from .recordCollection import RecordCollection

import metaknowledge

def diffusionGraph(source, target, weighted = True, sourceType = "raw", targetType = "raw", labelEdgesBy = None):
    """Takes in two [RecordCollections](../classes/RecordCollection.html#metaknowledge.RecordCollection) and produces a graph of the citations of _source_ by the [Records](../classes/Record.html#metaknowledge.Record) in _target_. By default the nodes in the are `Record` objects but this can be changed with the _sourceType_ and _targetType_ keywords. The edges of the graph go from the target to the source.

    Each node on the output graph has two boolean attributes, `"source"` and `"target"` indicating if they are targets or sources. Note, if the types of the sources and targets are different the attributes will not be checked for overlap of the other type. e.g. if the source type is `'TI'` (title) and the target type is `'UT'` (WOS number), and there is some overlap of the targets and sources. Then the Record corresponding to a source node will not be checked for being one of the titles of the targets, only its WOS number will be considered.

    # Parameters

    _source_ : `RecordCollection`

    > A metaknowledge `RecordCollection` containing the `Records` being cited

    _target_ : `RecordCollection`

    > A metaknowledge `RecordCollection` containing the `Records` citing those in _source_

    _weighted_ : `optional [bool]`

    > Default `True`, if `True` each edge will have an attribute `'weight'` giving the number of times the source has referenced the target.

    _sourceType_ : `optional [str]`

    > Default `'raw'`, if `'raw'` the returned graph will contain `Records` as source nodes.

    > If Records are not wanted then it can be set to a WOS tag, such as `'SO'` (for journals ), to make the nodes into the type of object returned by that tag from Records.

    _targetType_ : `optional [str]`

    > Default `'raw'`, if `'raw'` the returned graph will contain `Records` as target nodes.

    > If Records are not wanted then it can be set to a WOS tag, such as `'SO'` (for journals ), to make the nodes into the type of object returned by that tag from Records.

    _labelEdgesBy_ : `optional [str]`

    > Default `None`, if a WOS tag (or long name of WOS tag) then the edges of the output graph will have a attribute `'key'` that is the value of the referenced tag, of source `Record`, i.e. if `'PY'` is given then each edge will have a `'key'` value equal to the publication year of the source.

    > This option will cause the output graph to be an `MultiDiGraph` and is likely to result in parallel edges. If a `Record` has multiple values for at tag (e.g. `'AF'`) the each tag will create its own edge.

    # Returns

    `networkx Directed Graph or networkx multi Directed Graph`

    > A directed graph of the diffusion network, _labelEdgesBy_ is used the graph will allow parallel edges.
    """
    if sourceType != "raw" and sourceType not in tagsAndNameSet:
        raise RuntimeError("{} is not a valid node type, only 'raw' or those strings in tagsAndNameSet are allowed".format(sourceType))
    if targetType != "raw" and targetType not in tagsAndNameSet:
        raise RuntimeError("{} is not a valid node type, only 'raw' or those strings in tagsAndNameSet are allowed".format(targetType))
    if labelEdgesBy is not None:
        try:
            normVal = normalizeToTag(labelEdgesBy)
        except KeyError:
            raise RuntimeError ("{} is not a known tag, only tags in tagsAndNameSet are allowed.".format(labelEdgesBy))
        else:
            labelEdgesBy = normVal
    count = 0
    maxCount = len(source)
    progArgs = (0, "Starting to make a diffusion network")
    if metaknowledge.VERBOSE_MODE:
        progKwargs = {'dummy' : False}
    else:
        progKwargs = {'dummy' : True}
    with _ProgressBar(*progArgs, **progKwargs) as PBar:
        sourceDict = {}
        if labelEdgesBy is None:
            workingGraph = nx.DiGraph()
        else:
            workingGraph = nx.MultiDiGraph()
        for Rs in source:
            if PBar:
                count += 1
                PBar.updateVal(count / maxCount * .25, "Analyzing source: " + str(Rs))
            RsVal, RsExtras = makeNodeID(Rs, sourceType)
            if RsVal:
                sourceDict[Rs.createCitation()] = RsVal
                for val in RsVal:
                    if val not in workingGraph:
                        workingGraph.add_node(val, source = True, target = False, **RsExtras)
        if PBar:
            count = 0
            maxCount = len(target)
            PBar.updateVal(.25, "Done analyzing sources, starting on targets")
        for Rt in target:
            RtVal, RtExtras = makeNodeID(Rt, targetType)
            if labelEdgesBy is not None:
                edgeVals = Rt.get(labelEdgesBy)
                if edgeVals is None:
                    continue
                if not isinstance(edgeVals, list):
                    edgeVals = [edgeVals]
            if PBar:
                count += 1
                PBar.updateVal(count / maxCount * .75 + .25, "Analyzing target: " + str(Rt))
            if RtVal:
                for val in RtVal:
                    if val not in workingGraph:
                        workingGraph.add_node(val, source = False, target = True, **RtExtras)
                    else:
                        workingGraph.node[val]["target"] = True
                    targetCites = Rt.get('CR')
                    if targetCites:
                        for Rs in (sourceDict[c] for c in targetCites if c in sourceDict):
                            for sVal in Rs:
                                if labelEdgesBy is not None:
                                    for edgeVal in (str(ev) for ev in edgeVals):
                                        if weighted:
                                            if workingGraph.has_edge(sVal, val, key = edgeVal):
                                                for i, a in workingGraph[sVal][val].items():
                                                    if a['key'] == edgeVal:
                                                        workingGraph[sVal][val][i]['weight'] += 1
                                                        break
                                            else:
                                                attrDict = {'key' : edgeVal, 'weight' : 1}
                                                workingGraph.add_edge(sVal, val, attr_dict = attrDict)
                                        else:
                                            if not workingGraph.has_edge(sVal, val, key = edgeVal):
                                                workingGraph.add_edge(sVal, val, key = edgeVal)
                                else:
                                    if weighted:
                                        try:
                                            workingGraph[sVal][val]['weight'] += 1
                                        except KeyError:
                                            workingGraph.add_edge(sVal, val, weight = 1)
                                    else:
                                        workingGraph.add_edge(sVal, val)
        if PBar:
            PBar.finish("Done making a diffusion network of {} sources and {} targets".format(len(source), len(target)))
    return workingGraph

def diffusionCount(source, target, sourceType = "raw", extraValue = None, pandasFriendly = False,  compareCounts = False, numAuthors = True, useAllAuthors = True, _ProgBar = None, extraMapping = None):
    """Takes in two [RecordCollections](../classes/RecordCollection.html#metaknowledge.RecordCollection) and produces a `dict` counting the citations of _source_ by the [Records](../classes/Record.html#metaknowledge.Record) of _target_. By default the `dict` uses `Record` objects as keys but this can be changed with the _sourceType_ keyword to any of the WOS tags.

    # Parameters

    _source_ : `RecordCollection`

    > A metaknowledge `RecordCollection` containing the `Records` being cited

    _target_ : `RecordCollection`

    > A metaknowledge `RecordCollection` containing the `Records` citing those in _source_

    _sourceType_ : `optional [str]`

    > default `'raw'`, if `'raw'` the returned `dict` will contain `Records` as keys. If it is a WOS tag the keys will be of that type.

    _pandasFriendly_ : `optional [bool]`

    > default `False`, makes the output be a dict with two keys one `"Record"` is the list of Records ( or data type requested by _sourceType_) the other is their occurrence counts as `"Counts"`. The lists are the same length.

    _compareCounts_ : `optional [bool]`

    > default `False`, if `True` the diffusion analysis will be run twice, first with source and target setup like the default (global scope) then using only the source `RecordCollection` (local scope).

    _extraValue_ : `optional [str]`

    > default `None`, if a tag the returned dictionary will have `Records` mapped to maps, these maps will map the entries for the tag to counts. If _pandasFriendly_ is also `True` the resultant dictionary will have an additional column called `'year'`. This column will contain the year the citations occurred, in addition the Records entries will be duplicated for each year they occur in.

    > For example if `'year'` was given then the count for a single `Record` could be `{1990 : 1, 2000 : 5}`

    _useAllAuthors_ : `optional [bool]`

    > default `True`, if `False` only the first author will be used to generate the `Citations` for the _source_ `Records`

    # Returns

    `dict[:int]`

    > A dictionary with the type given by _sourceType_ as keys and integers as values.

    > If _compareCounts_ is `True` the values are tuples with the first integer being the diffusion in the target and the second the diffusion in the source.

    > If _pandasFriendly_ is `True` the returned dict has keys with the names of the WOS tags and lists with their values, i.e. a table with labeled columns. The counts are in the column named `"TargetCount"` and if _compareCounts_ the local count is in a column called `"SourceCount"`.
    """
    sourceCountString = "SourceCount"
    targetCountString = "TargetCount"
    if not isinstance(sourceType, str):
        raise RuntimeError("{} is not a valid node type, only tags or the string 'raw' are allowed".format(sourceType))
    if not isinstance(source, RecordCollection) or not isinstance(target, RecordCollection):
        raise RuntimeError("Source and target must be RecordCollections.")
    if extraValue is not None and not isinstance(extraValue, str):
        raise RuntimeError("{} is not a valid extraValue, only tags are allowed".format(extraValue))
    if extraMapping is None:
        extraMapping = lambda x : x
    if metaknowledge.VERBOSE_MODE or _ProgBar:
        if _ProgBar:
            PBar = _ProgBar
            PBar.updateVal(0, "Starting to analyse a diffusion network")
        else:
            PBar = _ProgressBar(0, "Starting to analyse a diffusion network")
        count = 0
        maxCount = len(source)
    else:
        PBar = _ProgressBar("Starting to analyse a diffusion network", dummy = True)
    count = 0
    maxCount = len(source)
    sourceDict = {}
    #Tells the function if the IDs are made of lists or of str
    listIds = None

    for Rs in source:
        if listIds is None and Rs.get(sourceType) is not None:
            listIds = isinstance(Rs.get(sourceType), list)
        count += 1
        PBar.updateVal(count / maxCount * .10, "Analyzing source: " + str(Rs))
        RsVal, RsExtras = makeNodeID(Rs, sourceType)
        if RsVal:
            if useAllAuthors:
                for c in Rs.createCitation(multiCite = True):
                    sourceDict[c] = RsVal
            else:
                sourceDict[Rs.createCitation()] = RsVal
    if extraValue is not None:
        if listIds:
            sourceCounts = {s : {targetCountString : 0} for s in itertools.chain.from_iterable(sourceDict.values())}
        else:
            sourceCounts = {s : {targetCountString : 0} for s in sourceDict.values()}
    else:
        if listIds:
            sourceCounts = {s : 0 for s in itertools.chain.from_iterable(sourceDict.values())}
        else:
            sourceCounts = {s : 0 for s in sourceDict.values()}
    count = 0
    maxCount = len(target)
    PBar.updateVal(.10, "Done analyzing sources, starting on targets")
    for Rt in target:
        count += 1
        PBar.updateVal(count / maxCount * .90 + .10, "Analyzing target: {}".format(Rt))
        targetCites = Rt.get('citations', [])
        if extraValue is not None:
            values = Rt.get(extraValue, [])
            if values is None:
                values = []
            elif not isinstance(values, list):
                values = [values]
            values = [extraMapping(val) for val in values]
        for c in  targetCites:
            try:
                RsourceVals = sourceDict[c]
            except KeyError:
                continue
            if listIds:
                for sVal in RsourceVals:
                    if extraValue:
                        sourceCounts[sVal][targetCountString] += 1
                        for val in values:
                            try:
                                sourceCounts[sVal][val] += 1
                            except KeyError:
                                sourceCounts[sVal][val] = 1
                    else:
                        sourceCounts[sVal] += 1
            else:
                if extraValue:
                    sourceCounts[RsourceVals][targetCountString] += 1
                    for val in values:
                        try:
                            sourceCounts[RsourceVals][val] += 1
                        except KeyError:
                            sourceCounts[RsourceVals][val] = 1
                else:
                    sourceCounts[RsourceVals] += 1
    if compareCounts:
        localCounts = diffusionCount(source, source, sourceType = sourceType, pandasFriendly = False,  compareCounts = False, extraValue = extraValue, _ProgBar = PBar)
    if PBar and not _ProgBar:
        PBar.finish("Done counting the diffusion of {} sources into {} targets".format(len(source), len(target)))
    if pandasFriendly:
        retDict = {targetCountString : []}
        if numAuthors:
            retDict["numAuthors"] = []
        if compareCounts:
            retDict[sourceCountString] = []
        if extraValue is not None:
            retDict[extraValue] = []
        if sourceType == 'raw':
            retrievedFields = []
            targetCount = []
            for R in sourceCounts.keys():
                tagsLst = [t for t in R.keys() if t not in retrievedFields]
                retrievedFields += tagsLst
            for tag in retrievedFields:
                retDict[tag] = []
            for R, occ in sourceCounts.items():
                if extraValue:
                    Rvals = R.subDict(retrievedFields)
                    for extraVal, occCount in occ.items():
                        retDict[extraValue].append(extraVal)
                        if numAuthors:
                            retDict["numAuthors"].append(len(R.get('authorsShort')))
                        for tag in retrievedFields:
                            retDict[tag].append(Rvals[tag])
                        retDict[targetCountString].append(occCount)
                        if compareCounts:
                            try:
                                retDict[sourceCountString].append(localCounts[R][extraVal])
                            except KeyError:
                                retDict[sourceCountString].append(0)
                else:
                    Rvals = R.subDict(retrievedFields)
                    if numAuthors:
                        retDict["numAuthors"].append(len(R.get('authorsShort')))
                    for tag in retrievedFields:
                        retDict[tag].append(Rvals[tag])
                    retDict[targetCountString].append(occ)
                    if compareCounts:
                        retDict[sourceCountString].append(localCounts[R])
        else:
            countLst = []
            recLst = []
            locLst = []
            if extraValue:
                extraValueLst = []
            for R, occ in sourceCounts.items():
                if extraValue:
                    for extraVal, occCount in occ.items():
                        countLst.append(occCount)
                        recLst.append(R)
                        extraValueLst.append(extraVal)
                        if compareCounts:
                            try:
                                locLst.append(localCounts[R][extraValue])
                            except KeyError:
                                locLst.append(0)
                else:
                    countLst.append(occ)
                    recLst.append(R)
                    if compareCounts:
                        locLst.append(localCounts[R])
            if compareCounts:
                retDict = {sourceType : recLst, targetCountString : countLst, sourceCountString : locLst}
            else:
                retDict = {sourceType : recLst, targetCountString : countLst}
            if extraValue:
                retDict[extraValue] = extraValueLst
        return retDict
    else:
        if compareCounts:
            for R, occ in localCounts.items():
                sourceCounts[R] = (sourceCounts[R], occ)
        return sourceCounts

def makeNodeID(Rec, ndType, extras = None):
    """Helper to make a node ID, extras is currently not used"""
    if ndType == 'raw':
        recID = Rec
    else:
        recID = Rec.get(ndType)
    if recID is None:
        pass
    elif isinstance(recID, list):
        recID = tuple(recID)
    else:
        recID = recID
    extraDict = {}
    if extras:
        for tag in extras:
            if tag == "raw":
                extraDict['Tag'] = Rec
            else:
                extraDict['Tag'] = Rec.get(tag)
    return recID, extraDict

def diffusionAddCountsFromSource(grph, source, target, nodeType = 'citations', extraType = None, diffusionLabel = 'DiffusionCount', extraKeys = None, countsDict = None, extraMapping = None):
    """Does a diffusion using [diffusionCount()](#metaknowledge.diffusion.diffusionCount) and updates _grph_ with it, using the nodes in the graph as keys in the diffusion, i.e. the source. The name of the attribute the counts are added to is given by _diffusionLabel_. If the graph is not composed of citations from the source and instead is another tag _nodeType_ needs to be given the tag string.

    # Parameters

    _grph_ : `networkx Graph`

    > The graph to be updated

    _source_ : `RecordCollection`

    > The `RecordCollection` that created _grph_

    _target_ : `RecordCollection`

    > The `RecordCollection` that will be counted

    _nodeType_ : `optional [str]`

    > default `'citations'`, the tag that constants the values used to create _grph_

    # Returns

    `dict[:int]`

    > The counts dictioanry used to add values to _grph_. *Note* _grph_ is modified by the function and the return is done in case you need it.
    """
    progArgs = (0, "Starting to add counts to graph")
    if metaknowledge.VERBOSE_MODE:
        progKwargs = {'dummy' : False}
    else:
        progKwargs = {'dummy' : True}
    with _ProgressBar(*progArgs, **progKwargs) as PBar:
        PBar.updateVal(0, 'Getting counts')
        if countsDict is None:
            countsDict = diffusionCount(source, target, sourceType = nodeType, extraValue = extraType, _ProgBar = PBar, extraMapping = extraMapping)
        try:
            if not isinstance(countsDict.keys().__iter__().__next__(), str):
                PBar.updateVal(.5, "Prepping the counts")
                newCountsDict = {}
                while True:
                    try:
                        k, v = countsDict.popitem()
                    except KeyError:
                        break
                    newCountsDict[str(k)] = v
                countsDict = newCountsDict
        except StopIteration:
            pass
        count = 0
        for n in grph.nodes_iter():
            PBar.updateVal(.5 + .5 * (count / len(grph)), "Adding count for '{}'".format(n))
            if extraType is not None:
                if extraKeys:
                    for key in extraKeys:
                        grph.node[n][key] = 0
                grph.node[n][diffusionLabel] = 0
                try:
                    for k, v in countsDict[n].items():
                        if k == 'TargetCount':
                            grph.node[n][diffusionLabel] = v
                        else:
                            if k:
                                grph.node[n][k] = v
                except KeyError:
                    grph.node[n][diffusionLabel] = 0
            else:
                grph.node[n][diffusionLabel] = countsDict.get(n, 0)
            count += 1
        PBar.finish("Done adding diffusion counts to a graph")
    return countsDict



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
from .metaknowledgeCLI import mkCLI
from .metaknowledgeMdToNb import mkMdToNb
from .metaknowledgeDocsGen import mkDocs



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
"""This is intended for metaknowledge only and may not work with anything else"""

import inspect
import argparse
import os
import time
import metaknowledge
import metaknowledge.WOS.tagProcessing
import importlib
import re

documentedModules = ['contour', 'WOS', 'medline', 'proquest', 'scopus', 'journalAbbreviations']

docsPrefix = time.strftime("%Y-%m-%d-")

funcCounter = 0
undocumented = 0

blurbDict = {
    #modules
    'contour' : "A nicer matplotlib graph visualizer and contour plot",
    'WOS' : "The backend functions and classes associated with the Web of Science",
    'journalAbbreviations' : "Handles the abbreviated journal names used by WOS",
    'medline' : "The backend functions and classes associated with Medline, the format used by Pubmed",
    'scopus' : "The backend functions and classes associated with records from scopus",
    'proquest' : "The backend functions and classes associated with ProQuest",

    #Classes

    'Citation' : "Citation are special, here is how they are handled",
    'WOSCitation' : "A Citation that supports the WOS journal abbreviations",

    'Record' : "The base of all the other Records, basically a dict",
    'ExtendedRecord' : "A Record the processes its contents before returning them",
    'WOSRecord' : "The object for containing and processing WOS entries",
    'ProQuestRecord' : "The object for containing and processing ProQuest entries",
    'MedlineRecord' : "The object for containing and processing Medline entries",
    'ScopusRecord' : "The object for containing and processing Scopus entries",

    'Grant' : "The base for all the other Grants",
    'FallbackGrant' : "The Grant used if a file was not identifiable",
    'CIHRGrant' : "The container for CIHR grant entries",
    'NSERCGrant' : "The container for NSERC grant entries",
    'MedlineGrant' : "The container for grants derived from Medline Records entries",
    'NSFGrant' : "The container for NSF grant entries",

    'Collection' : "The base of all other Collections, basically a set",
    'CollectionWithIDs' : "A Collection that only holds <i>metaknowledge</i> objects",
    'RecordCollection' : "A Collection of Records, this is what does most of the stuff on Records",
    'GrantCollection' : "A Collection of Grants, this is what does most of the stuff on Grants",

    #Deprecated
    'tagProcessing' : "All the tags and how they are handled",
}

singleFileYAML = """---
layout: page
title: Full Documentation {}
author:
- name: Reid McIlroy-Young
  department:
  affiliation: University of Waterloo, Waterloo ON, Canada
  email: rmcilroy@uwaterloo.ca
- name: John McLevey
  affiliation: University of Waterloo, Waterloo ON, Canada
  email: john.mclevey@uwaterloo.ca
shorttitle: metaknowledge
search_omit: true
---
""".format(metaknowledge.__version__)

def makeBlurb(name):
    if name in blurbDict:
        return blurbDict[name]
    else:
        print("\033[94m{} had no blurb\033[0m".format(name))
        return 'BLURB NEEDED FOR {}'.format(name)
        #raise RuntimeError("{} needs a blurb".format(name))

def makeHeader(title, excerpt, tags = (), weight = 10, layout = "doc", singleFile = False):
    if singleFile:
        s = """\n---\n<a name="{0}"></a>\n""".format(title)
    else:
        s = """---
layout: {4}
title: {0}
categories: docs
excerpt: {1}
tags: [{2}]
weight: {3}
---
<a name="{0}"></a>
""".format(title, excerpt, ', '.join(tags), weight, layout)
    return s

def argumentParser():
    parser = argparse.ArgumentParser(description="A simple script to generate docs for metaknowledge")
    parser.add_argument("-dir", "-d", default = os.path.normpath('.') ,nargs='?', help = 'Directory to write files to')
    parser.add_argument("-single", "-s", action = 'store_true', default = False, help = 'Write to only one file.')
    return parser.parse_args()

def getLineNumber(objTuple):
    try:
        ln = str(inspect.getsourcelines(objTuple[1])[1])
        fl = inspect.getfile(objTuple[1])
    except (OSError, TypeError):
        return '-1' #Because
    else:
        return fl + ln

def cleanargs(obj, basic = False, isClass = False):
    if isClass:
        argClasses = []
        for c in obj.__bases__:
            n = c.__name__
            if n in blurbDict:
                n = '<a href="#{0}"><u style="border-bottom: .5px dashed gray;">{0}</u></a>'.format(n)
            argClasses.append(n)
        argStr = ', '.join(argClasses)
    else:
        argStr = inspect.formatargspec(*inspect.getfullargspec(obj))
        argStr =  argStr.replace(", *args", '').replace(", **kwargs", '').replace('self, ', '').replace('self', '')[1:-1]
    if len(argStr) > 0:
        argStr = ', '.join([a for a in argStr.split(', ') if a[0] != '_'])
        if basic:
            return "(<i>{0}</i>)".format(argStr)
        else:
            return "(_{0}_)".format(argStr)
    else:
        return '()'

def makeUrls(s):
    return "[{0}]({{{{ site.baseurl }}}}{{{{ page.url }}}}#{3})".format(s.group(1), docsPrefix, s.group(2), s.group(3))

def makeSingleFileUrls(s):
    return "[{0}](#{1})".format(s.group(1), s.group(3))

def cleanedDoc(obj, lvl, singleFile = False):
    ds = inspect.getdoc(obj)
    if not isinstance(ds, str):
        raise AttributeError("'{}' on line {} of '{}' is missing its docstring".format(obj, inspect.getsourcelines(obj)[1], inspect.getfile(obj)))
    lns = ds.split('\n')
    nds = ''
    for line in lns:
        if singleFile:
            line = re.sub(r"\[(\S+)\]\(\#(\S+)\.(\S+)\)", makeSingleFileUrls, line, count = 99)
        else:
            line = re.sub(r"\[(\S+)\]\(\#(\S+)\.(\S+)\)", makeUrls, line, count = 99)
        if len(line) < 1:
            nds += '\n'
        elif line[0] == '#':
            nds += '#' * (lvl + 1) + line[1:] + '\n'
        elif line[0:4] == '    ':
            nds += '    ' + line[4:] + '\n'
        elif line[0] == '>':
            if line[1] != '>':
                nds += line[1:] + '\n'
            else:
                nds += "> " +  line[2:] + '\n'
        else:
            nds += line + '\n'
    return '{}\n\n'.format(nds)

def makeTitle(module, name, args = '', singleFile = False):
    if singleFile:
        s = '<a name="{1}"></a><small>{0}</small>**[<ins>{1}</ins>](#{1})**{2}:\n\n'.format(module, name, args)
    else:
        s = '<a name="{1}"></a><small>{0}</small>**[<ins>{1}</ins>]({{{{ site.baseurl }}}}{{{{ page.url }}}}#{1})**{2}:\n\n'.format(module, name, args)
    return s

def makeLine():
    style = [
    "padding: 0;",
    "border: none;",
    "border-width: 3px;",
    "height: 20px;",
    "color: #333;",
    "text-align: center;",
    "border-top-style: solid;",
    "border-bottom-style: solid;",
    ]
    return '<hr style="{}">'.format(''.join(style))

def makeTable(entries, header = '', prefix = '', withBlurbs = False, bigTable = False, simple = False, areClasses = False):
    ents = []
    if prefix:
        prefix = prefix + '.'
    for e in entries:
        if withBlurbs:
            if areClasses:
                ents.append("""<li><article><a href="#{0}"><b>{0}</b></a>{1}<span class="excerpt">{2}</span></article></li>""".format(e[0], cleanargs(e[1], basic = True, isClass = areClasses), makeBlurb(e[0])))
            else:
                ents.append("""<li><article><a href="#{0}"><b>{0}</b><span class="excerpt">{1}</span></a></article></li>""".format(e, makeBlurb(e)))
        elif bigTable:
            if e[2] is None:
                ents.append("""<li><article><a href="#{0}"><b>{0}</b>{1}</a></article></li>""".format(e[0], cleanargs(e[1], basic = True, isClass = areClasses), prefix))
            else:
                ents.append("""<li><article><a href="#{0}"><small>{2}</small>.<b>{0}</b>{1}</a></article></li>""".format(e[0], cleanargs(e[1], basic = True, isClass = areClasses), e[2]))
        elif simple:
            ents.append("""<li><article><b>{0}</b>{1}</article></li>""".format(e[0], cleanargs(e[1], isClass = areClasses, basic = True)))
        else:
            ents.append("""<li><article><a href="#{0}"><b>{0}</b>{1}</a></article></li>""".format(e[0], cleanargs(e[1], basic = True, isClass = areClasses), prefix))
    s = """<h3>{}</h3>\n\n<ol class="post-list">\n{}\n</ol>\n""".format(header,'\n'.join(ents))
    return s

def writeFunc(fn, f, prefix = '', level = 5, singleFile = False):
    global funcCounter
    funcCounter += 1
    f.write(makeLine() + "\n\n")
    f.write(makeTitle(prefix, fn[0], cleanargs(fn[1]), singleFile = singleFile))
    try:
        f.write(cleanedDoc(fn[1], lvl = level, singleFile = singleFile))
    except AttributeError:
        f.write("# Needs to be written\n\n")
        global undocumented
        undocumented += 1
        print("\033[93m{0}{1} had no docs\033[0m".format(prefix, fn[0]))

def writeClass(cl, f, prefix = '', level = 4, singleFile = False, exceptMode = False):
    f.write(makeTitle(prefix, cl[0], cleanargs(cl[1], isClass = True), singleFile = singleFile))
    if not exceptMode:
        f.write(makeTitle(prefix, "{}.__init__".format(cl[0]), cleanargs(cl[1].__init__), singleFile = singleFile))
    try:
        f.write(cleanedDoc(cl[1], lvl = level, singleFile = singleFile))
    except AttributeError:
        f.write("# Needs to be written\n\n")
        global undocumented
        undocumented += 1
        print("\033[93m{0}{1} had no docs\033[0m".format(prefix, cl[0]))

def proccessClass(cls, f, singleFile = False, exceptMode = False):
    writeClass(cls, f, singleFile = singleFile, exceptMode = exceptMode)
    baseMems = [ i for c in cls[1].__bases__ for i in inspect.getmembers(c)]
    if singleFile:
        f.write(makeLine())
    try:
        documented = cls[1]._documented
    except AttributeError:
        documented = []
    funcs = []
    for m in sorted(inspect.getmembers(cls[1]), key = getLineNumber):
        if m[0][0] == '_' or m in baseMems or m[0] == 'with_traceback':
            if m[0] in documented:
                funcs.append(m)
        elif inspect.isfunction(m[1]):
            funcs.append(m)
    if len(funcs) > 0 and not exceptMode:
        f.write(makeTable(funcs, prefix = cls[0], header = "\nThe {} class has the following methods:".format(cls[0])))
        for m in funcs:
            writeFunc(m, f, prefix = '{}.'.format(cls[0], singleFile = singleFile))

def writeClassFile(name, typ, targetFile = None, singleFile = False):
    fname = docsPrefix + "{}.md".format(name)
    if targetFile is not None:
        f = targetFile
    else:
        f = open(fname, 'w')
    f.write(makeHeader(name, makeBlurb(name), tags = ["class"], weight = 2, singleFile = singleFile))
    proccessClass((name, typ), f, singleFile = singleFile)
    if targetFile is None:
        f.write("\n{% include docsFooter.md %}")
        f.close()

def writeModuleFile(mod, targetFile = None, singleFile = False):
    print("Starting {}".format(mod))
    fname = docsPrefix + "{}.md".format(mod)
    if targetFile is not None:
        f = targetFile
    else:
        f = open(fname, 'w')
    f.write(makeHeader(mod, makeBlurb(mod), tags = ["module"], weight = 3, singleFile = singleFile))
    f.write('\n# [{0}]({{{{ site.baseurl }}}}{{{{ page.url }}}}#{0})\n\n'.format(mod))
    module = importlib.import_module('metaknowledge.{}'.format(mod))
    f.write(cleanedDoc(module, 3, singleFile = singleFile) + '\n\n')
    funcs = []
    for m in sorted(inspect.getmembers(module, predicate = inspect.isfunction), key = getLineNumber):
        if inspect.isbuiltin(m[1]) or m[0][0] == '_':
            pass
        elif inspect.isfunction(m[1]):
            funcs.append(m)
    if mod != "tagProcessing":
        f.write(makeTable(funcs, prefix = mod, header = '<a name="{0}">The <a href="#{0}"><u>{0}</u></a> module provides the following functions:</a>'.format(mod)))
        for fn in funcs:
            writeFunc(fn, f, prefix = "{}.".format(mod))
    else:
        for fn in metaknowledge.WOS.tagProcessing.tagToFunc.items():
            writeFunc((metaknowledge.tagToFull(fn[0]), fn[1]), f, prefix = "{}.".format(mod))
    if targetFile is None:
        f.write("\n{% include docsFooter.md %}")
        f.close()

def writeMainBody(funcs, vrs, exceptions, targetFile = None, singleFile = False):
    if targetFile:
        f = targetFile
    else:
        f = open(docsPrefix + "overview.md", 'w')

    f.write(makeHeader("Overview", "The <i>metaknowledge</i> Package, a quick tour", tags = ["main"], weight = 0, layout = "doc", singleFile = singleFile))
    f.write(cleanedDoc(metaknowledge, 3, singleFile = singleFile) + '\n\n')
    if targetFile is None:
        f.write("\n{% include docsFooter.md %}")
        f.close()
        f = open(docsPrefix + "metaknowledge.md", 'w')
    f.write(makeHeader("Base Functions", "The <i>metaknowledge</i> functions, for filtering reading and writing graphs", tags = ["functions"], weight = 1, layout = "doc", singleFile = singleFile))
    f.write(makeTable(funcs, header = "The functions provided by <i>metaknowledge</i> are:"))
    f.write(makeTable(exceptions, header = "The Exceptions defined by <i>metaknowledge</i> are:", simple = True, areClasses = True))
    for fnc in funcs:
        writeFunc(fnc, f)
    first = True
    if targetFile is None:
        f.write("\n{% include docsFooter.md %}")
        f.close()

def main(args):
    wDir = os.path.expanduser(os.path.normpath(args.dir))
    if not os.path.isdir(wDir):
        try:
            os.mkdir(wDir)
        except OSError:
            print('Creating the directory {} failed'.format(wDir))
            return 1
    os.chdir(wDir)

    classes = []
    funcs = []
    vrs = []
    exceptions = []
    builtins = []

    for m in sorted(inspect.getmembers(metaknowledge), key = getLineNumber):
        if inspect.isbuiltin(m[1]) or m[0][0] == '_':
            builtins.append(m)
        elif inspect.isclass(m[1]):
            if issubclass(m[1], Exception):
                exceptions.append(m)
            else:
                classes.append(m)
        elif inspect.isfunction(m[1]):
            funcs.append(m)
        else:
            vrs.append(m)

    if args.single:
        single = True
        f = open("metaknowledgeFull.md",'w')
        f.write(singleFileYAML)

        f.write(makeTable(documentedModules, header = '<a name="objlist"></a>The modules of <i>metaknowledge</i> are:', withBlurbs = True))

        f.write(makeTable(classes, header = '<a name="objlist"></a>The classes of <i>metaknowledge</i> are:', withBlurbs = True, areClasses = True))

        bigTableEntries = [(f[0], f[1], None) for f in funcs]

        for cls in classes:
            baseMems = inspect.getmembers(cls[1].__bases__[0])
            for m in sorted(inspect.getmembers(cls[1]), key = getLineNumber):
                if m[0][0] == '_' or m in baseMems:
                    pass
                elif inspect.isfunction(m[1]):
                    bigTableEntries.append((m[0], m[1], cls[0]))

        f.write(makeTable(bigTableEntries, header = '<a name="fulllist"></a>All the functions and methods of <i>metaknowledge</i> and its objects are as follows:', bigTable = True))

        for mod in documentedModules:
            modTableEntries = []
            module = importlib.import_module('metaknowledge.{}'.format(mod))
            for m in sorted(inspect.getmembers(module, predicate = inspect.isfunction), key = getLineNumber):
                if inspect.isbuiltin(m[1]) or m[0][0] == '_':
                    pass
                elif inspect.isfunction(m[1]):
                    modTableEntries.append((m[0], m[1], mod))
            f.write(makeTable(modTableEntries, header = 'All the functions of the <a href="#{0}"><u>{0}</u></a> module are as follows:'.format(mod), bigTable = True))

    else:
        single = False
        f = None

    writeMainBody(funcs, vrs, exceptions, targetFile = f, singleFile = single)

    for cls in classes:
        writeClassFile(*cls, targetFile = f, singleFile = single)

    for mod in documentedModules:
        writeModuleFile(mod, targetFile = f, singleFile = single)

    if args.single:
        f.write("\n{% include docsFooter.md %}")
        f.close

def mkDocs():
    args = argumentParser()
    main(args)
    print("{} total functions".format(funcCounter))
    print("{} undocumented".format(undocumented))

if __name__ == '__main__':
    mkDocs()



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
import unittest
import shutil
import os

import metaknowledge

class TestGrantCollection(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        metaknowledge.VERBOSE_MODE = False
        cls.GCmain = metaknowledge.GrantCollection("metaknowledge/tests/", cached = True)

    def setUp(self):
        self.GC = self.GCmain.copy()

    def test_empty(self):
        GCempty = metaknowledge.GrantCollection()
        self.assertEqual(len(GCempty), 0)
        self.assertEqual(GCempty.name, "Empty")

    def test_creationErrors(self):
        with self.assertRaises(metaknowledge.mkExceptions.GrantCollectionException):
            GCbad = metaknowledge.GrantCollection("README.md", extension = '.csv')
        with self.assertRaises(metaknowledge.mkExceptions.BadInputFile):
            GCbad = metaknowledge.GrantCollection("README.md")
        with self.assertRaises(metaknowledge.mkExceptions.BadInputFile):
            GCbad = metaknowledge.GrantCollection("README.md", extension = '.md')
        with self.assertRaises(metaknowledge.mkExceptions.BadInputFile):
            GCbad = metaknowledge.GrantCollection(".", extension = '.md')
        with self.assertRaises(metaknowledge.mkExceptions.GrantCollectionException):
            GCbad = metaknowledge.GrantCollection("README")
        with self.assertRaises(metaknowledge.mkExceptions.GrantCollectionException):
            GCbad = metaknowledge.GrantCollection(1)
        with self.assertRaises(metaknowledge.mkExceptions.GrantCollectionException):
            GCbad = metaknowledge.GrantCollection({1})

    def test_creation(self):
        self.assertIsInstance(self.GC, metaknowledge.GrantCollection)
        self.assertIsInstance(self.GC, metaknowledge.Collection)
        self.assertAlmostEqual(len(self.GC), 2022, delta = 4)
        self.assertIsInstance(self.GC.peek(), metaknowledge.Record)
        self.assertEqual(metaknowledge.GrantCollection(self.GC), self.GC)

    def test_Caching(self):
        self.assertTrue(os.path.isfile("metaknowledge/tests/tests.[].mkGrantDirCache"))
        os.remove("metaknowledge/tests/tests.[].mkGrantDirCache")

    def test_fallback(self):
        fname = "DefaultGrantTestFile.csv"
        shutil.copyfile("metaknowledge/tests/NSERC_TEST_PARTNER.testcsv", fname)
        GC = metaknowledge.GrantCollection(fname, extension = '.csv')
        self.assertEqual(GC._collectedTypes, {"FallbackGrant"})
        os.remove(fname)

    def test_CoInstitution(self):
        G = self.GC.networkCoInvestigatorInstitution()
        self.assertEqual(metaknowledge.graphStats(G), 'Nodes: 641\nEdges: 2034\nIsolates: 79\nSelf loops: 0\nDensity: 0.00991615\nTransitivity: 0.273548')



from ..mkExceptions import BadProQuestFile

from .recordProQuest import ProQuestRecord

def isProQuestFile(infile, checkedLines = 2):
    """Determines if _infile_ is the path to a ProQuest file. A file is considered to be a Proquest file if it has the correct encoding (`utf-8`) and within the first _checkedLines_ the following starts.

        ____________________________________________________________

        Report Information from ProQuest

    # Parameters

    _infile_ : `str`

    > The path to the targets file

    _checkedLines_ : `optional [int]`

    > default 2, the number of lines to check for the header

    # Returns

    `bool`

    > `True` if the file is a valid ProQuest file
    """
    try:
        with open(infile, 'r', encoding='utf-8') as openfile:
            f = enumerate(openfile, start = 0)
            for i in range(checkedLines):
                #This seems like enough checking
                #The next line is the date meaning it is not constant
                #More checking could be done
                if f.__next__()[1] == "_" * 60 + "\n" and f.__next__()[1] == '\n' and f.__next__()[1] == 'Report Information from ProQuest\n':
                    return True
    except (StopIteration, UnicodeDecodeError):
        return False
    else:
        return False

def proQuestParser(proFile):
    """Parses a ProQuest file, _proFile_, to extract the individual entries.

    A ProQuest file has three sections, first a list of the contained entries, second the full metadata and finally a bibtex formatted entry for the record. This parser only uses the first two as the bibtex contains no information the second section does not. Also, the first section is only used to verify the second section. The returned [ProQuestRecord](../classes/ProQuestRecord.html#metaknowledge.proquest.ProQuestRecord) contains the data from the second section, with the same key strings as ProQuest uses and the unlabeled sections are called in order, `'Name'`, `'Author'` and `'url'`.

    # Parameters

    _proFile_ : `str`

    > A path to a valid ProQuest file, use [isProQuestFile](#metaknowledge.proquest.proQuestHandlers.isProQuestFile) to verify

    # Returns

    `set[ProQuestRecord]`

    > Records for each of the entries
    """
    #assumes the file is ProQuest
    nameDict = {}
    recSet = set()
    error = None
    lineNum = 0
    try:
        with open(proFile, 'r', encoding = 'utf-8') as openfile:
            f = enumerate(openfile, start = 1)
            for i in range(12):
                lineNum, line = next(f)
            # f is file so it *should* end, or at least cause a parser error eventually
            while True:
                lineNum, line = next(f)
                lineNum, line = next(f)
                if line == 'Bibliography\n':
                    for i in range(3):
                        lineNum, line = next(f)
                    break
                else:
                    s = line.split('. ')
                    nameDict[int(s[0])] = '. '.join(s[1:])[:-1]
            while True:
                #import pdb; pdb.set_trace()
                lineNum, line = next(f)
                if line == 'Bibliography\n':
                    break
                elif line.startswith('Document '):
                    n = int(line[9:].split(' of ')[0])
                    R = ProQuestRecord(f, sFile = proFile, sLine = lineNum)
                    if R.get('Title') != nameDict[n]:
                        error = BadProQuestFile("The numbering of the titles at the beginning of the file does not match the records inside. Line {} has a record titled '{}' with number {}, the name should be '{}'.".format(lineNum, R.get('Title', "TITLE MISSING"), n, nameDict[n]))
                        raise StopIteration
                    recSet.add(R)
                    lineNum, line = next(f)
                else:
                    #Parsing failed
                    error = BadProQuestFile("The file '{}' has parts of it that are unparsable starting at line: {}. It is likely that the seperators between the records are incorrect".format(proFile, lineNum))
                    raise StopIteration
    except (UnicodeDecodeError, StopIteration, ValueError) as e:
        if error is None:
            error = BadProQuestFile("The file '{}' has parts of it that are unparsable starting at line: {}.\nThe error was: '{}'".format(proFile, lineNum, e))
    return recSet, error
